from typing import Union

from sirius_sdk.messaging import Message


class BaseConfidentialStorageError(RuntimeError):

    PROBLEM_CODE = 'confidential_storage_error'

    def __init__(self, message, *args):
        super().__init__(message, *args)

    @property
    def message(self) -> str:
        return self.args[0] if self.args else ''


class ConfidentialStorageUnexpectedMessageType(BaseConfidentialStorageError):

    PROBLEM_CODE = 'invalid_message_type'

    def __init__(self, message: Union[str, Message], *args, **kwargs):
        if isinstance(message, Message):
            if '@type' in message:
                typ = message.get('@type')
                err_msg = f'Unexpected message type: "{typ}"'
            else:
                err_msg = f'Unexpected message type of type: "{message.__class__.__name__}"'
            super().__init__(err_msg)
        else:
            super().__init__(message)


class ConfidentialStorageInvalidRequest(BaseConfidentialStorageError):
    PROBLEM_CODE = 'invalid_request'


class StreamEOF(BaseConfidentialStorageError):
    PROBLEM_CODE = 'eof'


class EncryptionError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'encryption_error'


class StreamInitializationError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'initialization_error'


class StreamSeekableError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'stream_is_not_seekable'


class DocumentFormatError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'document_format_error'


class StreamFormatError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'stream_format_error'


class ConfidentialStorageTimeoutOccurred(BaseConfidentialStorageError):
    PROBLEM_CODE = 'timeout_occurred'


class ConfidentialStoragePermissionDenied(BaseConfidentialStorageError):
    PROBLEM_CODE = 'permission_denied'


class DataVaultCreateResourceError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'resource_error'


class DataVaultResourceMissing(BaseConfidentialStorageError):
    PROBLEM_CODE = 'resource_missing'


class DataVaultSessionError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'session_error'


class DataVaultStateError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'state_error'


class DataVaultOSError(BaseConfidentialStorageError):
    PROBLEM_CODE = 'os_error'

    def __init__(self, message: Union[OSError, str], *args):
        if isinstance(message, OSError):
            err_msg = str(message)
            for arg in message.args:
                if isinstance(arg, str):
                    err_msg = arg
                    break
        else:
            err_msg = message
        super().__init__(err_msg, *args)



from sirius_sdk.agent.agent import Agent
from sirius_sdk.hub import init, context, endpoints, ledger, dkms, subscribe, ping, send, send_to, send_batched, \
    generate_qr_code, DID, Crypto, Microledgers, PairwiseList, AnonCreds, CoProtocolThreadedP2P, CoProtocolP2PAnon, \
    CoProtocolP2P, AbstractP2PCoProtocol, CoProtocolThreadedTheirs, Cache, open_communication, NonSecrets, \
    acquire, release, Config, spawn_coprotocol, prepare_response
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.abstract.p2p import Endpoint, TheirEndpoint, Pairwise
from sirius_sdk.agent import aries_rfc
from sirius_sdk import recipes
from sirius_sdk.agent import didcomm
from sirius_sdk.agent.dkms import Schema, CredentialDefinition, AnonCredSchema, Ledger, DKMS, NYMRole
from sirius_sdk.errors import indy_exceptions, exceptions
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.abstract.api import APICrypto
from sirius_sdk.abstract.batching import RoutingBatch
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets, RetrieveRecordOptions as NonSecretsRetrieveRecordOptions
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList


__all__ = [
    "Agent", "P2PConnection", "init", "context", "endpoints", "ledger", "dkms",
    "subscribe", "ping", "send", "send_to", "send_batched", "generate_qr_code", "DID", "Crypto", "Microledgers", "PairwiseList",
    "CoProtocolThreadedP2P", "CoProtocolP2PAnon", "CoProtocolP2P", "AbstractP2PCoProtocol", "Pairwise",
    "aries_rfc", "CoProtocolThreadedTheirs", "AnonCreds", "Cache", "open_communication", "prepare_response", "NonSecrets", "APICrypto",
    "Schema", "CredentialDefinition", "AnonCredSchema", "indy_exceptions", "exceptions", "Ledger", "acquire", "release",
    "Config", "didcomm", "recipes", "AbstractPairwiseList", "AbstractNonSecrets", "RoutingBatch", "NYMRole",
    "AbstractCache", "AbstractDID", "AbstractMicroledgerList", "DKMS", "spawn_coprotocol", "NonSecretsRetrieveRecordOptions"
]



import datetime
import json
import uuid

import pytest

import sirius_sdk
from sirius_sdk import Agent
from sirius_sdk.base import WebSocketConnector, Message
from sirius_sdk.errors.exceptions import SiriusTimeoutIO
from tests.helpers import ServerTestSuite
from tests.conftest import get_pairwise


@pytest.mark.asyncio
async def test_agent_rcv_timeout(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent4')
    timeout = 3
    # Check-1: check timeout error if ttl limit was set globally
    conn_with_global_setting = WebSocketConnector(
        server_address=params['server_address'],
        path='/rpc',
        credentials=params['credentials'],
        timeout=timeout,
    )
    await conn_with_global_setting.open()
    try:
        context = await conn_with_global_setting.read()
        assert context
        with pytest.raises(SiriusTimeoutIO):
            await conn_with_global_setting.read()
    finally:
        await conn_with_global_setting.close()

    # Check-2: check timeout error if ttl limit was set locally
    conn_with_local_setting = WebSocketConnector(
        server_address=params['server_address'],
        path='/rpc',
        credentials=params['credentials'],
        timeout=10000,  # infinite
    )
    await conn_with_local_setting.open()
    try:
        context = await conn_with_local_setting.read()
        assert context
        with pytest.raises(SiriusTimeoutIO):
            await conn_with_local_setting.read(timeout)
    finally:
        await conn_with_local_setting.close()

    # Check-3: check timeout for ttl was set greater than global setting
    conn_with_little_global_timeout = WebSocketConnector(
        server_address=params['server_address'],
        path='/rpc',
        credentials=params['credentials'],
        timeout=1,  # low value
    )
    await conn_with_little_global_timeout.open()
    try:
        context = await conn_with_little_global_timeout.read()
        assert context
        stamp1 = datetime.datetime.utcnow()
        with pytest.raises(SiriusTimeoutIO):
            await conn_with_little_global_timeout.read(timeout=5)
        stamp2 = datetime.datetime.utcnow()
        stamps_delta = stamp2 - stamp1
        assert 4 <= stamps_delta.total_seconds() <= 6, f'Timeout {stamps_delta.total_seconds()}'
    finally:
        await conn_with_little_global_timeout.close()


@pytest.mark.skip
@pytest.mark.asyncio
async def test_state_machines_timeout(test_suite: ServerTestSuite):
    params_me = test_suite.get_agent_params('agent4')
    params_their = test_suite.get_agent_params('agent3')
    timeout = 5
    me = Agent(
        server_address=params_me['server_address'],
        credentials=params_me['credentials'],
        p2p=params_me['p2p'],
        timeout=1000,
        name='agent4'
    )
    their = Agent(
        server_address=params_their['server_address'],
        credentials=params_their['credentials'],
        p2p=params_their['p2p'],
        timeout=timeout,
        name='agent3'
    )
    await me.open()
    await their.open()
    try:
        p2p = await get_pairwise(me, their)
        their_conn_key = await their.wallet.crypto.create_key()
        their_endpoint = [e for e in their.endpoints if e.routing_keys == []][0]
    finally:
        await me.close()
        await their.close()

    stamp1 = datetime.datetime.utcnow()
    async with sirius_sdk.context(params_me['server_address'], params_me['credentials'], params_me['p2p']):
        endpoints = await sirius_sdk.endpoints()
        endpoint = [e for e in endpoints if e.routing_keys == []][0]
        rfc_0160 = sirius_sdk.aries_rfc.Invitee(me=p2p.me, my_endpoint=endpoint, time_to_live=timeout)
        success, p2p = await rfc_0160.create_connection(
            invitation=sirius_sdk.aries_rfc.Invitation(
                label='Their', recipient_keys=[their_conn_key], endpoint=their_endpoint.address
            ),
            my_label='Me'
        )
        assert success is False
        assert rfc_0160.problem_report is not None
    stamp2 = datetime.datetime.utcnow()
    stamps_delta = stamp2 - stamp1
    assert 4 <= stamps_delta.total_seconds() <= 6, f'Timeout {stamps_delta.total_seconds()}'


@pytest.mark.asyncio
async def test_echo_big_message(agent4: Agent):
    await agent4.open()
    try:
        lst = [str(x) for x in range(100000)]
        big_data = ''.join(lst)
        stamp1 = datetime.datetime.utcnow()
        ret = await agent4.echo(message=big_data)
        stamp2 = datetime.datetime.utcnow()
        stamps_delta = stamp2 - stamp1
        stamps_delta_sec = stamps_delta.total_seconds()
        assert stamps_delta_sec < 1, f'Timeout {stamps_delta.total_seconds()}'
    finally:
        await agent4.close()



from typing import Any


class BaseSiriusException(Exception):
    
    def __init__(self, message: str = None, *args, **kwargs):
        super(BaseSiriusException, self).__init__(message, *args, **kwargs)
        self.message = message

    @staticmethod
    def _prefix_msg(msg, prefix=None):
        return "{}{}".format(
            "" if prefix is None else "{}: ".format(prefix),
            msg
        )

    def __str__(self):
        return f'{self.message}'


class SiriusConnectionClosed(BaseSiriusException):
    pass


class SiriusTimeoutIO(BaseSiriusException):
    pass


class SiriusRPCError(BaseSiriusException):
    pass


class SiriusTimeoutRPC(SiriusRPCError):
    pass


class SiriusIOError(BaseSiriusException):
    pass


class SiriusInvalidPayloadStructure(BaseSiriusException):
    pass


class SiriusUnsupportedData(BaseSiriusException):
    pass


class SiriusPendingOperation(BaseSiriusException):
    pass


class SiriusValueEmpty(BaseSiriusException):
    pass


class SiriusAlreadyTriggered(BaseSiriusException):
    pass


class SiriusPromiseContextException(BaseSiriusException):

    def __init__(self, class_name: str, printable: str, *args, **kwargs):
        self.class_name = class_name
        self.printable = printable
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f'class_name: {self.class_name}; printable: {self.printable}'


class SiriusCryptoError(BaseSiriusException):
    """ Failed crypto call. """


class SiriusInvalidMessage(BaseSiriusException):
    """ Thrown when message is malformed. """


class SiriusInvalidType(BaseSiriusException):
    """ When type is unparsable or invalid. """


class SiriusValidationError(BaseSiriusException):
    pass


class SiriusContextError(BaseSiriusException):
    pass


class SiriusInitializationError(BaseSiriusException):
    pass


class SiriusInvalidMessageClass(BaseSiriusException):
    pass


class SiriusFieldTypeError(BaseSiriusException, TypeError):
    """Exception for TypeError

    Extends TypeError to provide formatted error message

    :param v_name: variable name
    :param v_value: variable value
    :param v_exp_t: expected variable type
    """

    def __init__(self, v_name: str, v_value: Any, v_exp_t: Any, *args, prefix=None):
        super().__init__(
            self._prefix_msg(
                ("variable '{}', type {}, expected: {}".format(v_name, type(v_value), v_exp_t)),
                prefix
            ), *args
        )


class SiriusFieldValueError(BaseSiriusException, ValueError):
    """Exception for ValueError

    Extends ValueError to provide formatted error message

    :param v_name: variable name
    :param v_value: variable value
    :param v_exp_value: expected variable value
    :param prefix: (optional) prefix for the message
    """
    def __init__(self, v_name: str, v_value: Any, v_exp_value: Any, *args,
                 prefix=None):
        super().__init__(
            self._prefix_msg(
                ("variable '{}', value {}, expected: {}".format(v_name, v_value, v_exp_value)),
                prefix
            ), *args
        )


class StateMachineTerminatedWithError(BaseSiriusException):

    def __init__(self, problem_code: str, explain: str, notify: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.problem_code = problem_code
        self.explain = explain
        self.notify = notify

    def __str__(self):
        return f'problem_code: {self.problem_code}; explain: {self.explain}; notify: {self.notify}'


class OperationAbortedManually(BaseSiriusException):
    pass


class StateMachineAborted(BaseSiriusException):
    pass


class SiriusTransportError(BaseSiriusException):
    pass



from typing import Any, Optional, List

import pytest

import sirius_sdk
from sirius_sdk.abstract.bus import AbstractBus
from sirius_sdk.abstract.api import APICoProtocols
from sirius_sdk import APICrypto

from tests.helpers import ServerTestSuite


class OverriddenMethodCalled(RuntimeError):
    pass


class OverriddenBus(AbstractBus):

    async def subscribe(self, thid: str) -> bool:
        raise OverriddenMethodCalled

    async def subscribe_ext(self, sender_vk: List[str], recipient_vk: List[str], protocols: List[str]) -> (bool, List[str]):
        raise OverriddenMethodCalled

    async def unsubscribe(self, thid: str):
        raise OverriddenMethodCalled

    async def unsubscribe_ext(self, thids: List[str]):
        raise OverriddenMethodCalled

    async def publish(self, thid: str, payload: bytes) -> int:
        raise OverriddenMethodCalled

    async def publish_ext(self, binding_ids: List[str], payload: bytes) -> int:
        raise OverriddenMethodCalled

    async def get_event(self, timeout: float = None) -> AbstractBus.BytesEvent:
        raise OverriddenMethodCalled

    async def get_message(self, timeout: float = None) -> AbstractBus.MessageEvent:
        raise OverriddenMethodCalled

    async def abort(self):
        raise OverriddenMethodCalled


class OverriddenCoprotocols(APICoProtocols):

    async def spawn_coprotocol(self) -> AbstractBus:
        return OverriddenBus()


class OverriddenCrypto(APICrypto):

    async def create_key(self, seed: str = None, crypto_type: str = None) -> str:
        return 'KEY'

    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        raise NotImplemented

    async def get_key_metadata(self, verkey: str) -> Optional[dict]:
        raise NotImplemented

    async def crypto_sign(self, signer_vk: str, msg: bytes) -> bytes:
        raise NotImplemented

    async def crypto_verify(self, signer_vk: str, msg: bytes, signature: bytes) -> bool:
        raise NotImplemented

    async def anon_crypt(self, recipient_vk: str, msg: bytes) -> bytes:
        raise NotImplemented

    async def anon_decrypt(self, recipient_vk: str, encrypted_msg: bytes) -> bytes:
        raise NotImplemented

    async def pack_message(self, message: Any, recipient_verkeys: list, sender_verkey: str = None) -> bytes:
        raise NotImplemented

    async def unpack_message(self, jwe: bytes) -> dict:
        raise NotImplemented


@pytest.mark.asyncio
async def test_config_cloud(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent3')
    cfg = sirius_sdk.Config().setup_cloud(params['server_address'], params['credentials'], params['p2p'])
    async with sirius_sdk.context(cfg):
        test = await sirius_sdk.DID.list_my_dids_with_meta()
        assert type(test) is list


@pytest.mark.asyncio
async def test_override_crypto():
    cfg = sirius_sdk.Config().override_crypto(dependency=OverriddenCrypto())
    async with sirius_sdk.context(cfg):
        key = await sirius_sdk.Crypto.create_key()
        assert key == 'KEY'


@pytest.mark.asyncio
async def test_override_coprotocols():
    cfg = sirius_sdk.Config().override_coprotocols(dependency=OverriddenCoprotocols())
    async with sirius_sdk.context(cfg):
        bus = await sirius_sdk.spawn_coprotocol()
        with pytest.raises(OverriddenMethodCalled):
            await bus.subscribe('thid')



from typing import Optional

from sirius_sdk import Pairwise
from sirius_sdk.agent import Pong, Ping
from sirius_sdk.hub.coprotocols import CoProtocolThreadedP2P


async def ping_their(their: Pairwise, comment: str = None, wait_timeout: int = 15) -> (bool, Optional[Pong]):
    """Send pin g to remote participant and wait ong response"""
    ping = Ping(
        comment=comment,
        response_requested=True
    )
    co = CoProtocolThreadedP2P(
        thid=ping.id,
        to=their,
        time_to_live=wait_timeout
    )
    success, pong = await co.switch(ping)
    if success and isinstance(pong, Pong):
        return True, pong
    else:
        return False, None



from abc import ABC, abstractmethod
from typing import Optional, List, Any


class AbstractPairwise(ABC):

    @abstractmethod
    async def is_pairwise_exists(self, their_did: str) -> bool:
        """
        Check if pairwise is exists.

        :param their_did: encoded Did.
        :return: true - if pairwise is exists, false - otherwise
        """
        raise NotImplemented

    @abstractmethod
    async def create_pairwise(self, their_did: str, my_did: str, metadata: dict=None, tags: dict=None) -> None:
        """
        Creates pairwise.

        :param their_did: encrypting DID
        :param my_did: encrypting DID
        :param metadata: (Optional) extra information for pairwise
        :param tags: tags for searching operations
        :return: Error code
        """
        raise NotImplemented

    @abstractmethod
    async def list_pairwise(self) -> List[Any]:
        """
        Get list of saved pairwise.

        :return: pairwise_list: list of saved pairwise
        """
        raise NotImplemented

    @abstractmethod
    async def get_pairwise(self, their_did: str) -> Optional[dict]:
        """
        Gets pairwise information for specific their_did.

        :param their_did: encoded Did
        :return: pairwise_info_json: did info associated with their did
        """
        raise NotImplemented

    @abstractmethod
    async def set_pairwise_metadata(self, their_did: str, metadata: dict=None, tags: dict=None) -> None:
        """
        Save some data in the Wallet for pairwise associated with Did.

        :param their_did: encoded DID
        :param metadata: some extra information for pairwise
        :param tags: tags for searching operation
        """
        raise NotImplemented

    @abstractmethod
    async def search(self, tags: dict, limit: int = None) -> (List[dict], int):
        """Search Pairwises

        :param tags: tags based query
        :param limit: max items count
        :return: Results, TotalCount
        """
        raise NotImplemented



import datetime
from enum import Enum
from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import List, Union, Optional, Set

from sirius_sdk.abstract.p2p import Pairwise
from sirius_sdk.encryption.ed25519 import ensure_is_bytes
from .streams import AbstractReadOnlyStream, AbstractWriteOnlyStream, BaseStreamEncryption, \
    ReadOnlyStreamDecodingWrapper, WriteOnlyStreamEncodingWrapper, StreamEncryption, StreamDecryption
from .documents import EncryptedDocument
from .errors import ConfidentialStoragePermissionDenied
from .encoding import JWE, KeyPair
from .utils import datetime_to_utc_str


@dataclass
class HMAC:
    # An identifier for the HMAC key. The value is required a MUST be or map to a URI.
    id: str
    # The type of HMAC key. The value is required and MUST be or map to a URI.
    type: str

    @property
    def is_filled(self) -> bool:
        if self.id or self.type:
            return True
        else:
            return False

    def as_json(self) -> dict:
        return {'id': self.id, 'type': self.type}


class DocumentMeta(dict):

    def __init__(self, created: Union[str, datetime.datetime] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if created is not None:
            if isinstance(created, datetime.datetime):
                self['created'] = datetime_to_utc_str(created)
            else:
                self['created'] = created


class StreamMeta(DocumentMeta):

    def __init__(
            self, created: Union[str, datetime.datetime] = None,
            chunks: int = None,
            content_type: str = None,  # "video/mpeg", "image/png"
            *args, **kwargs
    ):
        super().__init__(created, *args, **kwargs)
        if chunks is not None:
            self['chunks'] = chunks
        if content_type is not None:
            self['contentType'] = content_type


class DataVaultStreamWrapper:

    def __init__(self, readable: AbstractReadOnlyStream = None, writable: AbstractWriteOnlyStream = None):
        self._readable = readable
        self._writable = writable

    async def readable(self, jwe: Union[JWE, dict] = None, keys: KeyPair = None) -> AbstractReadOnlyStream:
        if jwe is None:
            return self._readable
        else:
            enc = StreamDecryption.from_jwe(jwe)
            if keys is not None:
                enc.setup(vk=keys.pk, sk=keys.sk)
            return ReadOnlyStreamDecodingWrapper(src=self._readable, enc=enc)

    async def writable(self, jwe: Union[JWE, dict] = None, cek: Union[bytes, str] = None) -> AbstractWriteOnlyStream:
        if jwe is None:
            return self._writable
        else:
            if isinstance(cek, str):
                cek = ensure_is_bytes(cek)
            return WriteOnlyStreamEncodingWrapper(dest=self._writable, enc=StreamEncryption.from_jwe(jwe, cek))


class StructuredDocument:
    """A structured document is used to store application data as well as metadata about the application data.
       This information is typically encrypted and then stored on the data vault.

    see details:  https://identity.foundation/confidential-storage/#structureddocument
    """
    @dataclass
    class Index:
        sequence: int = None
        hmac: HMAC = None
        attributes: List[str] = None

    def __init__(
            self, id_: str, meta: dict,
            urn: str = None, indexed: List[Index] = None,
            content: EncryptedDocument = None,
            stream: DataVaultStreamWrapper = None
    ):
        self.__id = id_
        self.__meta = dict(**meta)
        self.__urn = urn
        if isinstance(content, AbstractReadOnlyStream):
            self.__meta['chunks'] = content.chunks_num
        self.__content = content
        self.__stream = stream
        self.__indexed: List[StructuredDocument.Index] = indexed or []

    @property
    def id(self) -> str:
        return self.__id

    @property
    def urn(self) -> str:
        return self.__urn or 'urn:id:' + self.__id

    @property
    def meta(self) -> dict:
        return self.__meta

    @property
    def content(self) -> Optional[EncryptedDocument]:
        return self.__content

    @property
    def doc(self) -> Optional[EncryptedDocument]:
        return self.__content

    @doc.setter
    def doc(self, value: Optional[EncryptedDocument]):
        self.__content = value

    @property
    def stream(self) -> Optional[DataVaultStreamWrapper]:
        return self.__stream

    @stream.setter
    def stream(self, value: Optional[DataVaultStreamWrapper]):
        self.__stream = value

    @property
    def indexed(self) -> List['Index']:
        return self.__indexed


class ConfidentialStorageAuthProvider:
    """Layer C: Authorization (policy enforcement point)

    see details: https://identity.foundation/confidential-storage/#ecosystem-overview
    """

    class PermissionLevel(Enum):
        # Entity may run read-operations
        CAN_READ = 'CAN_READ'
        # Entity has access to write/update document/streams
        CAN_WRITE = 'CAN_WRITE'
        # Entity can create new documents/streams
        CAN_CREATE = 'CAN_CREATE'
        # Entity can update resource metadata
        CAN_UPDATE = 'CAN_UPDATE'

    def __init__(self):
        self.__entity = None
        self.__authorized = False

    @property
    def entity(self) -> Optional[Pairwise]:
        return self.__entity

    @property
    def authorized(self) -> bool:
        return self.__authorized

    async def authorize(self, entity: Pairwise) -> bool:
        """Authorize participant

        In DESCENDANTS may invoke adv. proof verify procedures...

        :param entity: Service Client
        :return: success
        """
        self.__entity = entity
        self.__authorized = True
        return self.__authorized

    def has_permissions(self) -> Set[PermissionLevel]:
        """
        :return: list of DataVault permissions for client
        """
        return {
            self.PermissionLevel.CAN_READ,
            self.PermissionLevel.CAN_WRITE,
            self.PermissionLevel.CAN_CREATE,
            self.PermissionLevel.CAN_UPDATE
        }

    @property
    def can_read(self) -> bool:
        return self.PermissionLevel.CAN_READ in self.has_permissions()

    @property
    def can_write(self) -> bool:
        return self.PermissionLevel.CAN_WRITE in self.has_permissions()

    @property
    def can_create(self) -> bool:
        return self.PermissionLevel.CAN_CREATE in self.has_permissions()

    @property
    def can_update(self) -> bool:
        return self.PermissionLevel.CAN_UPDATE in self.has_permissions()

    def validate(
            self, can_read: bool = False, can_write: bool = False,
            can_create: bool = False, can_update: bool = False, expected: Set[PermissionLevel] = None
    ):
        expected = list(expected) if expected else []
        permissions = self.has_permissions()
        if can_read:
            expected.append(ConfidentialStorageAuthProvider.PermissionLevel.CAN_READ)
        if can_write:
            expected.append(ConfidentialStorageAuthProvider.PermissionLevel.CAN_WRITE)
        if can_create:
            expected.append(ConfidentialStorageAuthProvider.PermissionLevel.CAN_CREATE)
        if can_update:
            expected.append(ConfidentialStorageAuthProvider.PermissionLevel.CAN_UPDATE)
        if not set(expected).issubset(permissions):
            expected_str = ','.join([p.value for p in expected])
            raise ConfidentialStoragePermissionDenied(f'Expected permissions: [{expected_str}]')


@dataclass
class VaultConfig:
    """Data vault configuration isn't strictly necessary for using the other features of data vaults.
    This should have its own conformance section/class or potentially event be non-normative.

    details: https://identity.foundation/confidential-storage/#datavaultconfiguration"""

    @dataclass
    class KeyAgreement:
        # An identifier for the key agreement key. The value is required and MUST be a URI.
        # The key agreement key is used to derive a secret that is then used
        # to generate a key encryption key for the receiver.
        id: str
        # The type of key agreement key. The value is required and MUST be or map to a URI.
        type: str = 'X25519KeyAgreementKey2019'

        @property
        def is_filled(self) -> bool:
            if self.id or self.type:
                return True
            else:
                return False
    id: str
    # The entity or cryptographic key that is in control of the data vault.
    # The value is required and MUST be a URI. Example: "did:example:123456789"
    controller: str
    # A unique counter for the data vault in order to ensure that clients are
    # properly synchronized to the data vault. Example: 0
    sequence: int = 0
    # The root entities or cryptographic key(s) that are authorized to invoke an authorization capability
    # to modify the data vault's configuration or read or write to it.
    # The value is optional, but if present, MUST be a URI or an array of URIs.
    # When this value is not present, the value of controller property is used for the same purpose.
    invoker: Optional[Union[str, List[str]]] = None
    # The root entities or cryptographic key(s) that are authorized to delegate authorization capabilities
    # to modify the data vault's configuration or read or write to it.
    # The value is optional, but if present, MUST be a URI or an array of URIs.
    # When this value is not present, the value of controller property is used for the same purpose.
    delegator: Optional[Union[str, List[str]]] = None
    # Used to express an application-specific reference identifier.
    # The value is optional and, if present, MUST be a string.
    reference_id: Optional[str] = None
    # keyAgreementKey
    key_agreement: Optional[KeyAgreement] = None
    # HMAC
    hmac: Optional[HMAC] = None

    def as_json(self) -> dict:
        js = {
            'id': self.id,
            'sequence': self.sequence,
            'controller': self.controller,
        }
        if self.invoker:
            js['invoker'] = self.invoker
        if self.delegator:
            js['delegator'] = self.delegator
        if self.reference_id:
            js['referenceId'] = self.reference_id
        if self.key_agreement and self.key_agreement.is_filled:
            js['keyAgreementKey'] = {
                'id': self.key_agreement.id,
                'type': self.key_agreement.type
            }
        if self.hmac and self.hmac.is_filled:
            js['hmac'] = {
                'id': self.hmac.id,
                'type': self.hmac.type
            }
        return js

    def from_json(self, js: dict):
        js = dict(**js)
        self.id = js.get('id', None)
        self.sequence = js.pop('sequence', 0)
        self.controller = js.pop('controller', None)
        self.invoker = js.pop('invoker', None)
        self.reference_id = js.pop('referenceId', None)
        key_agreement = js.pop('keyAgreementKey', {})
        key_agreement_id = key_agreement.get('id', None)
        key_agreement_type = key_agreement.get('type', None)
        if key_agreement_id or key_agreement_type:
            self.key_agreement = VaultConfig.KeyAgreement(
                key_agreement_id, key_agreement_type
            )
        else:
            self.key_agreement = None
        hmac = js.pop('hmac', {})
        hmac_id = hmac.get('id', None)
        hmac_type = hmac.get('type', None)
        if hmac_id or hmac_type:
            self.hmac = HMAC(
                hmac_id, hmac_type
            )
        else:
            self.hmac = None
        # Copy others
        for fld, value in js.items():
            self.__setattr__(fld, value)

    @staticmethod
    def create_from_json(js: dict) -> 'VaultConfig':
        inst = VaultConfig(id='id', controller='')
        inst.from_json(js)
        return inst


class ConfidentialStorageRawByteStorage(ABC):
    """Layer A: raw bytes storage (Cloud, DB, File-system, Mobile, etc)

    see details: https://identity.foundation/confidential-storage/#ecosystem-overview
    """

    def __init__(self, encryption: BaseStreamEncryption = None):
        self.__encryption = encryption

    @property
    def encryption(self) -> Optional[BaseStreamEncryption]:
        return self.__encryption

    @abstractmethod
    async def create(self, uri: str):
        raise NotImplementedError

    @abstractmethod
    async def remove(self, uri: str):
        raise NotImplementedError

    @abstractmethod
    async def readable(self, uri: str, chunks_num: int) -> AbstractReadOnlyStream:
        raise NotImplementedError

    @abstractmethod
    async def writeable(self, uri: str) -> AbstractWriteOnlyStream:
        raise NotImplementedError

    @abstractmethod
    async def exists(self, uri: str) -> bool:
        raise NotImplementedError


class EncryptedDataVault:
    """Layer B: Encrypted Vault Storage

    see details: https://identity.foundation/confidential-storage/#ecosystem-overview
    """

    def __init__(self, auth: ConfidentialStorageAuthProvider = None, cfg: VaultConfig = None):
        if auth is not None and auth.authorized is False:
            raise ConfidentialStoragePermissionDenied(f'Not authorized access')
        self.__auth = auth
        self._cfg = cfg
        if cfg is None and auth is not None and auth.authorized:
            their_did = auth.entity.their.did
            my_did = auth.entity.me.did
            my_vk = auth.entity.me.verkey
            if ':' not in their_did:
                their_did = f'did:peer:{their_did}'
            if ':' not in my_did:
                my_did = f'did:peer:{my_did}'
            if ':' not in my_vk:
                my_vk = f'did:key:{my_vk}'
            cfg = VaultConfig(
                id=f'did:edvs:{auth.entity.their.did}',
                sequence=0,
                controller=my_did,
                delegator=[their_did],
                key_agreement=VaultConfig.KeyAgreement(
                    id=my_vk,
                    type='X25519KeyAgreementKey2019'
                ),
                reference_id='default'
            )
        self._cfg = cfg

    @property
    def auth(self) -> ConfidentialStorageAuthProvider:
        return self.__auth

    @property
    def cfg(self) -> VaultConfig:
        return self._cfg

    class Indexes:

        @abstractmethod
        async def filter(self, **attributes) -> List[StructuredDocument]:
            raise NotImplementedError

    @abstractmethod
    async def open(self):
        raise NotImplementedError

    @abstractmethod
    async def close(self):
        raise NotImplementedError

    @abstractmethod
    async def indexes(self) -> Indexes:
        raise NotImplementedError

    @abstractmethod
    async def create_stream(self, uri: str, meta: Union[dict, StreamMeta] = None, chunk_size: int = None, **attributes) -> StructuredDocument:
        raise NotImplementedError

    @abstractmethod
    async def create_document(self, uri: str, meta: Union[dict, DocumentMeta] = None, **attributes) -> StructuredDocument:
        raise NotImplementedError

    @abstractmethod
    async def remove(self, uri: str):
        raise NotImplementedError

    @abstractmethod
    async def update(self, uri: str, meta: Union[dict, DocumentMeta, StreamMeta] = None, **attributes):
        raise NotImplementedError

    @abstractmethod
    async def load(self, uri: str) -> StructuredDocument:
        raise NotImplementedError

    @abstractmethod
    async def save_document(self, uri: str, doc: EncryptedDocument):
        raise NotImplementedError

    @abstractmethod
    async def readable(self, uri: str) -> AbstractReadOnlyStream:
        raise NotImplementedError

    @abstractmethod
    async def writable(self, uri: str) -> AbstractWriteOnlyStream:
        raise NotImplementedError



import json
import asyncio
import logging
import math
import hashlib
from typing import Union, Optional, List, Dict

import aiohttp

import sirius_sdk
import sirius_sdk.abstract.p2p
from sirius_sdk.abstract.listener import AbstractListener, Event
from sirius_sdk.didcomm import extentions as didcomm_ext
from sirius_sdk.base import BaseConnector, INFINITE_TIMEOUT
from sirius_sdk.abstract.p2p import Endpoint
from sirius_sdk.hub.coprotocols import AbstractP2PCoProtocol
from sirius_sdk.abstract.api import APIRouter, APICoProtocols
from sirius_sdk.errors.exceptions import *
from sirius_sdk.abstract.bus import AbstractBus
from sirius_sdk.messaging import Message, restore_message_instance, Type as MsgType
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.agent.aries_rfc.did_doc import DIDDoc
from sirius_sdk.agent.aries_rfc.mixins import ThreadMixin as AriesThreadMixin
from sirius_sdk.agent.aries_rfc.feature_0753_bus.messages import *
from sirius_sdk.agent.aries_rfc.feature_0212_pickup.messages import *


def qualify_key(key: str) -> str:
    if ':' not in key:
        return f'did:key:{key}'
    else:
        return key


class MediatorConnector(BaseConnector):

    IO_TIMEOUT = 60
    ENC = 'utf-8'

    def __init__(
            self, uri: str, loop: asyncio.AbstractEventLoop = None, timeout: int = IO_TIMEOUT,
            http_headers: dict = None
    ):
        self.__session = aiohttp.ClientSession(
            loop=loop,
            timeout=aiohttp.ClientTimeout(total=timeout),
            headers=http_headers
        )
        self._uri = uri
        self._ws = None
        self.__timeout = timeout

    def __del__(self):
        asyncio.ensure_future(self.__session.close())

    @property
    def is_open(self):
        return self._ws is not None and not self._ws.closed

    async def open(self):
        if not self.is_open:
            self._ws = await self.__session.ws_connect(url=self._uri, ssl=False)

    async def close(self):
        if self.is_open:
            await self._ws.close()
            self._ws = None

    async def read(self, timeout: float = None) -> bytes:
        if timeout == INFINITE_TIMEOUT:
            _timeout = None
        else:
            _timeout = timeout or self.__timeout
        try:
            msg = await self._ws.receive(timeout=_timeout)
        except asyncio.TimeoutError as e:
            raise SiriusTimeoutIO() from e
        if msg.type in [aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSED]:
            raise SiriusConnectionClosed()
        elif msg.type == aiohttp.WSMsgType.TEXT:
            return msg.data.encode(self.ENC)
        elif msg.type == aiohttp.WSMsgType.BINARY:
            return msg.data
        elif msg.type == aiohttp.WSMsgType.ERROR:
            raise SiriusIOError()

    async def write(self, message: Union[Message, bytes]) -> bool:
        if isinstance(message, Message):
            payload = message.serialize().encode(self.ENC)
        else:
            payload = message
        await self._ws.send_bytes(payload)
        return True


class TunnelMixin:

    async def pack(self, message: Message) -> bytes:
        packed = await sirius_sdk.Crypto.pack_message(
            message=json.dumps(message),
            recipient_verkeys=[self.mediator_verkey],
            sender_verkey=self.my_verkey,
        )
        return packed

    async def unpack(self, jwe: bytes) -> (Optional[Union[Message, bytes]], str, Optional[str]):

        if b'protected' in jwe:
            decrypted = await sirius_sdk.Crypto.unpack_message(jwe)
            message, sender_vk, recip_vk = decrypted['message'], decrypted['sender_verkey'], decrypted['recipient_verkey']
        else:
            message, sender_vk, recip_vk = json.loads(jwe.decode()), None, None

        if type(message) is str:
            payload = json.loads(message)
        elif type(message) is bytes:
            payload = json.loads(message.decode())
        elif type(message) is dict:
            payload = message
        else:
            raise SiriusInvalidMessage('Unexpected message type')
        if 'protected' in payload:
            return message.encode(), sender_vk, recip_vk
        else:
            success, msg = restore_message_instance(payload)
            if success:
                return msg, sender_vk, recip_vk
            else:
                return Message(**payload), sender_vk, recip_vk


class MediatorCoProtocol(TunnelMixin, AbstractP2PCoProtocol):

    def __init__(
            self,
            connector: MediatorConnector, my_verkey: str, mediator_verkey: str,
            time_to_live: int = None
    ):
        super().__init__(time_to_live)
        self.__connector: MediatorConnector = connector
        self._mediator_verkey = mediator_verkey
        self._my_verkey = my_verkey

    @property
    def connector(self) -> MediatorConnector:
        return self.__connector

    @property
    def mediator_verkey(self) -> str:
        return self._mediator_verkey

    @property
    def my_verkey(self) -> str:
        return self._my_verkey

    async def send(self, message: Message):
        # Aries-RFC 0092 https://github.com/hyperledger/aries-rfcs/tree/master/features/0092-transport-return-route
        transport = message.get('~transport', {})
        transport['return_route'] = 'all'
        message['~transport'] = transport
        payload = await self.pack(message)
        await self.connector.write(payload)

    async def get_one(self, timeout: int = None) -> (Optional[Message], str, Optional[str]):
        payload = await self.connector.read(timeout)
        message, sender_vk, recip_vk = await self.unpack(payload)
        return message, sender_vk, recip_vk

    async def switch(self, message: Message) -> (bool, Message):
        await self.send(message)
        msg, _, _ = await self.get_one()
        return True, msg


class MediatorBus(AbstractBus, TunnelMixin):

    BINDINGS_CTX_ID = 'binding.id'

    def __init__(self, connector: MediatorConnector, my_verkey: str, mediator_verkey: str):
        self.__connector: MediatorConnector = connector
        self._mediator_verkey = mediator_verkey
        self._my_verkey = my_verkey
        self.__client_id = str(id(self))
        self.__binding_ids: Dict[str, str] = {}

    @property
    def connector(self) -> MediatorConnector:
        return self.__connector

    @property
    def mediator_verkey(self) -> str:
        return self._mediator_verkey

    @property
    def my_verkey(self) -> str:
        return self._my_verkey

    async def subscribe(self, thid: str) -> bool:
        request = BusSubscribeRequest(cast=BusSubscribeRequest.Cast(thid=thid), parent_thread_id=self.__client_id)
        payload = await self.pack(request)
        await self.connector.write(payload)
        jwe = await self.connector.read()
        resp, _, _ = await self.unpack(jwe)
        self.__validate(resp, expected_class=BusBindResponse)
        if resp.thread_id != thid:
            self.__set_binding_id(thid, resp.binding_id)
        return True

    async def subscribe_ext(self, sender_vk: List[str], recipient_vk: List[str], protocols: List[str]) -> (bool, List[str]):
        binding_id = []
        for _sender_vk in sender_vk:
            for _recipient_vk in recipient_vk:
                for protocol in protocols:
                    thid = self.__binding_id_from_attrs(_sender_vk, _recipient_vk, protocol)
                    success = await self.subscribe(thid)
                    if success:
                        binding_id.append(thid)
        return len(binding_id) > 0, binding_id

    async def unsubscribe(self, thid: str):
        binding_id = self.__pop_binding_id(thid)
        request = BusUnsubscribeRequest(
            thread_id=binding_id or thid,
            need_answer=False,  # don't wait response
        )
        payload = await self.pack(request)
        await self.connector.write(payload)

    async def unsubscribe_ext(self, thids: List[str]):
        request = BusUnsubscribeRequest(
            thread_id=thids,
            need_answer=False,  # don't wait response
            parent_thread_id=self.__client_id
        )
        payload = await self.pack(request)
        await self.connector.write(payload)

    async def publish(self, thid: str, payload: bytes) -> int:
        binding_id = self.__get_binding_id(thid)
        request = BusPublishRequest(thread_id=binding_id or thid, payload=payload)
        payload = await self.pack(request)
        await self.connector.write(payload)
        jwe = await self.connector.read()
        resp, _, _ = await self.unpack(jwe)
        self.__validate(resp, expected_class=BusPublishResponse)
        return resp.recipients_num

    async def get_event(self, timeout: float = None) -> AbstractBus.BytesEvent:
        expire_at = datetime.datetime.now() + datetime.timedelta(seconds=timeout) if timeout is not None else None

        def __in_loop__() -> bool:
            if expire_at is None:
                return True
            else:
                return datetime.datetime.now() <= expire_at

        while __in_loop__():
            if expire_at is None:
                wait_timeout = None
            else:
                dt_diff = expire_at - datetime.datetime.now()
                wait_timeout = dt_diff.total_seconds()

            request = PickUpNoop()
            if wait_timeout is not None:
                request.timing = PickUpNoop.Timing(delay_milli=math.ceil(wait_timeout * 1000))
            else:
                request.timing = PickUpNoop.Timing(delay_milli=5*60 * 1000)  # wait for 5 min by default

            request.please_ack = True
            payload = await self.pack(request)
            await self.connector.write(payload)
            if timeout is None:
                read_timeout = None
            else:
                read_timeout = 1.1*timeout   # Increase if pickup protocol not raise response
            jwe = await self.connector.read(read_timeout)

            resp, sender_vk, recip_vk = await self.unpack(jwe)
            if isinstance(resp, Message):
                thread = ThreadMixin.get_thread(resp)
                if thread and ((thread.pthid == self.__client_id) or (thread.thid == request.id)):
                    if isinstance(resp, BusEvent):
                        return AbstractBus.BytesEvent(thread_id=resp.thread_id, payload=resp.payload)
                    elif isinstance(resp, BusBindResponse):
                        if resp.aborted is True and resp.parent_thread_id == self.__client_id:
                            raise OperationAbortedManually('Bus events awaiting was aborted by user')
                    elif isinstance(resp, BusProblemReport):
                        raise SiriusRPCError(resp.explain)
                    elif isinstance(resp, PickUpProblemReport):
                        if resp.problem_code == PickUpProblemReport.PROBLEM_CODE_TIMEOUT_OCCURRED:
                            if wait_timeout is not None:
                                raise SiriusTimeoutIO
                            else:
                                continue
                        else:
                            raise SiriusRPCError(resp.explain)
                else:
                    logging.warning(
                        f'Bus listener: message was ignored cause of unexpected thread binging, '
                        f'expected pthid: {self.__client_id}'
                    )
                    logging.warning(json.dumps(resp, indent=2, sort_keys=True))
            if isinstance(resp, bytes):
                logging.critical('Unexpected bytes received')
                # fwd, fwd_sender_vk, fwd_recip_vk = await self.unpack(resp)
                # await self.__raise_bus_message(fwd, resp, fwd_sender_vk, fwd_recip_vk)

    async def get_message(self, timeout: float = None) -> AbstractBus.MessageEvent:
        event = await self.get_event(timeout)
        decrypted = await sirius_sdk.Crypto.unpack_message(event.payload)
        message = decrypted['message']
        if isinstance(message, str):
            message = json.loads(message)
        if message.get('@type', None):
            ok, msg = restore_message_instance(decrypted)
            if not ok:
                msg = Message(**message)
            return AbstractBus.MessageEvent(
                thread_id=event.thread_id,
                message=msg,
                sender_verkey=decrypted.get('sender_verkey', None),
                recipient_verkey=decrypted.get('recipient_verkey', None)
            )
        else:
            raise SiriusRPCError('Unexpected message format')

    async def abort(self):
        request = BusUnsubscribeRequest(parent_thread_id=self.__client_id, aborted=True)
        payload = await self.pack(request)
        await self.connector.write(payload)

    @staticmethod
    def __validate(msg: BusOperation, expected_class) -> BusOperation:
        if isinstance(msg, BusProblemReport):
            raise SiriusRPCError(msg.explain)
        elif isinstance(msg, expected_class):
            return msg
        else:
            raise SiriusRPCError(f'Unexpected response type: {msg.__class__.__name__}')

    def __get_binding_id(self, thid: str) -> Optional[str]:
        return self.__binding_ids.get(thid, None)

    def __pop_binding_id(self, thid: str) -> Optional[str]:
        bid = self.__binding_ids.get(thid, None)
        if bid is not None:
            del self.__binding_ids[thid]
        return bid

    def __set_binding_id(self, thid: str, binding_id: str):
        self.__binding_ids[thid] = binding_id

    async def __raise_bus_message(self, decrypted: Message, jwe: bytes, sender_vk, recip_vk):
        if '@type' in decrypted:
            thread = AriesThreadMixin.get_thread(decrypted)
            if thread and thread.thid:
                num = await self.publish(thread.thid, jwe)
                print('')

    @staticmethod
    def __binding_id_from_attrs(sender_vk: Optional[str], recipient_vk: Optional[str], protocol: str) -> str:
        return f'protocol:{sender_vk}/{recipient_vk}/{protocol}'


class MediatorListener(TunnelMixin, AbstractListener):

    def __init__(
            self, connector: MediatorConnector, my_verkey: Optional[str],
            mediator_verkey: Optional[str], pairwise_resolver: AbstractPairwiseList = None
    ):
        self.__connector: MediatorConnector = connector
        self.__pairwise_resolver = pairwise_resolver
        self._mediator_verkey = mediator_verkey
        self._my_verkey = my_verkey

    @property
    def connector(self) -> MediatorConnector:
        return self.__connector

    @property
    def mediator_verkey(self) -> str:
        return self._mediator_verkey

    @property
    def my_verkey(self) -> str:
        return self._my_verkey

    def is_open(self) -> bool:
        return self.__connector.is_open

    async def get_one(self, timeout: int = None) -> Event:
        if timeout is None:
            wait_timeout = INFINITE_TIMEOUT
        else:
            wait_timeout = timeout
        payload = await self.connector.read(timeout=wait_timeout)
        message, sender_vk, recip_vk = await self.unpack(payload)
        if sender_vk is not None and self.__pairwise_resolver is not None:
            p2p = await self.__pairwise_resolver.load_for_verkey(sender_vk)
        else:
            p2p = None
        kwargs = {
           '@type': 'https://didcomm.org/sirius_rpc/1.0/event',
           'message': message,
           'recipient_verkey': recip_vk,
           'sender_verkey': sender_vk,
           'jwe': payload
        }
        event = Event(pairwise=p2p, **kwargs)
        return event


class Mediator(APIRouter, APICoProtocols):

    FIREBASE_SERVICE_TYPE = 'FCMService'
    DEFAULT_GROUP_ID = 'default'
    # Reserved group-id
    COPROTOCOLS_GROUP_ID = 'coprotocols_group_id_for_decrypt'

    def __init__(
            self, uri: str, my_verkey: str, mediator_verkey: str,
            firebase_device_id: str = None,
            mediator_label: str = None,
            routing_keys: List[str] = None,
            timeout: int = MediatorConnector.IO_TIMEOUT,
            pairwise_resolver: AbstractPairwiseList = None
    ):
        self.__uri = uri
        self.__timeout = timeout
        self.__pairwise_resolver = pairwise_resolver
        self._connector = MediatorConnector(
            uri=self.__uri, timeout=self.__timeout,
            http_headers={
                didcomm_ext.return_route.HEADER_NAME: didcomm_ext.return_route.RouteType.ALL.value
            }
        )
        self.__my_verkey = my_verkey
        self.__mediator_verkey = mediator_verkey
        self.__mediator_label = mediator_label
        self._coprotocol = MediatorCoProtocol(
            connector=self._connector,
            my_verkey=self.__my_verkey,
            mediator_verkey=self.__mediator_verkey,
            time_to_live=self.__timeout
        )
        my_did_bytes = sirius_sdk.encryption.did_from_verkey(
            sirius_sdk.encryption.b58_to_bytes(my_verkey)
        )
        self.me = sirius_sdk.abstract.p2p.Pairwise.Me(
            did=sirius_sdk.encryption.bytes_to_b58(my_did_bytes),
            verkey=my_verkey
        )
        self._my_did = sirius_sdk.encryption.did_from_verkey(
            sirius_sdk.encryption.b58_to_bytes(my_verkey)
        )
        self.firebase_device_id = firebase_device_id
        self._mediator_invitation = sirius_sdk.aries_rfc.Invitation(
            label=self.__mediator_label or 'Mediator',
            recipient_keys=[self.__mediator_verkey],
            endpoint=uri
        )
        self._endpoints = []
        self.__is_connected = False
        self.__did_doc = None
        if routing_keys is None:
            routing_keys = []
        self.__routing_keys = [qualify_key(key) for key in routing_keys]
        self.__endpoints: List[sirius_sdk.abstract.p2p.Endpoint] = []
        self.__bus: Optional[AbstractBus] = None

    def copy(self) -> "Mediator":
        inst = Mediator(
            uri=self.__uri, my_verkey=self.__my_verkey, mediator_verkey=self.__mediator_verkey,
            mediator_label=self.__mediator_label, routing_keys=self.__routing_keys, timeout=self.__timeout,
            pairwise_resolver=self.__pairwise_resolver
        )
        return inst

    @property
    def is_connected(self) -> bool:
        return self.__is_connected

    @property
    def did_doc(self) -> Optional[dict]:
        return self.__did_doc

    @property
    def endpoints(self) -> List[sirius_sdk.abstract.p2p.Endpoint]:
        return self.__endpoints

    @property
    def bus(self) -> AbstractBus:
        return self.__bus

    async def connect(self):
        if self.__is_connected:
            return
        # Run P2P connection establishment according Aries-RFC0160
        if not self._connector.is_open:
            await self._connector.open()
        try:
            self.__is_connected = True
            success, mediator_did_doc = await self.__connect_to_mediator(
                endpoint=didcomm_ext.return_route.URI_QUEUE_TRANSPORT,
                firebase_device_id=self.firebase_device_id,
                group_id='off'
            )
            if success:
                # 3. P2P successfully established
                self.__did_doc = mediator_did_doc
                if self.__routing_keys:
                    # 4. Actualize routing_keys
                    keys_request = sirius_sdk.aries_rfc.KeylistQuery()
                    success, keys_response = await self._coprotocol.switch(keys_request)
                    if success:
                        mediator_mediate_keys = keys_response['keys']
                        keys_to_add = [key for key in self.__routing_keys if key not in mediator_mediate_keys]
                        command = sirius_sdk.aries_rfc.KeylistUpdate(
                            updates=[
                                sirius_sdk.aries_rfc.KeylistAddAction(recipient_key=key)
                                for key in keys_to_add
                            ]
                        )
                        success, update_response = await self._coprotocol.switch(command)
                        if not success:
                            raise SiriusRPCError('Error while updating mediator keys list')
                    else:
                        raise SiriusRPCError('Error while requesting mediator keys list')
                # Final: grant endpoint
                mediate_request = sirius_sdk.aries_rfc.MediateRequest()
                success, mediate_grant = await self._coprotocol.switch(mediate_request)
                if success:
                    routing_keys = self.__routing_keys
                    routing_keys.extend(mediate_grant['routing_keys'])
                    self.__endpoints.append(
                        sirius_sdk.abstract.p2p.Endpoint(
                            address=mediate_grant['endpoint'],
                            routing_keys=routing_keys,
                            is_default=True
                        )
                    )
                    self.__bus = MediatorBus(
                        connector=self._connector,
                        my_verkey=self._coprotocol.my_verkey,
                        mediator_verkey=self._coprotocol.mediator_verkey
                    )
                else:
                    raise SiriusRPCError('Error while granting mediate endpoint')
            else:
                raise SiriusConnectionClosed()
        except BaseSiriusException:
            await self.disconnect()
            raise

    async def disconnect(self):
        if self.__is_connected:
            await self._connector.close()
            self.__is_connected = False
            self.__did_doc = None
            self.__bus = None

    async def get_endpoints(self) -> List[Endpoint]:
        return self.__endpoints

    async def subscribe(self, group_id: str = None) -> AbstractListener:
        # Re-Open RPC
        await self._connector.close()
        await self._connector.open()
        # Redeclare Group-ID in DIDDoc to re-schedule downstream
        success, diddoc = await self.__connect_to_mediator(
            endpoint='ws://',
            group_id=group_id or self.DEFAULT_GROUP_ID
        )
        if success:
            diddoc = DIDDoc(diddoc)
            mediator_services = diddoc.extract_service(high_priority=True, type_='MediatorService')
            uri = mediator_services['serviceEndpoint']
            conn = MediatorConnector(uri)
            await conn.open()
            listener = MediatorListener(
                connector=conn,
                my_verkey=None,
                mediator_verkey=None
            )
            return listener
        else:
            raise SiriusRPCError('Error while configure load-balancer')

    async def spawn_coprotocol(self) -> AbstractBus:
        bus = MediatorBus(
            connector=self._connector,
            my_verkey=self._coprotocol.my_verkey,
            mediator_verkey=self._coprotocol.mediator_verkey
        )
        return bus

    async def __connect_to_mediator(
            self, endpoint: str = didcomm_ext.return_route.URI_QUEUE_TRANSPORT,
            firebase_device_id: str = None, group_id: str = None
    ) -> (bool, Optional[dict]):
        # Run P2P connection establishment according Aries-RFC0160
        # - RFC: https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol
        # - recipient declare endpoint address as "ws://" that means communication is established over duplex channel
        #   see details: https://github.com/hyperledger/aries-rfcs/tree/main/features/0092-transport-return-route
        state_machine = sirius_sdk.aries_rfc.Invitee(
            me=self.me,
            my_endpoint=sirius_sdk.abstract.p2p.Endpoint(
                address=didcomm_ext.return_route.URI_QUEUE_TRANSPORT,
                routing_keys=[]
            ),
            coprotocol=self._coprotocol
        )
        # 1. Recipient DIDDoc contains Firebase device id inside service with type "FCMService"
        did_doc = sirius_sdk.aries_rfc.ConnRequest.build_did_doc(
            did=self.me.did,
            verkey=self.me.verkey,
            endpoint=endpoint
        )
        if group_id is not None:
            services_num = len(did_doc['service'])
            for n in range(services_num):
                did_doc['service'][n]['group_id'] = group_id
        did_doc_extra = {'service': did_doc['service']}
        if firebase_device_id:
            did_doc_extra['service'].append({
                "id": 'did:peer:' + self.me.did + ";indy",
                "type": self.FIREBASE_SERVICE_TYPE,
                "recipientKeys": [],
                "priority": 1,
                "serviceEndpoint": firebase_device_id
            })
        success, p2p = await state_machine.create_connection(
            invitation=self._mediator_invitation,
            my_label=f'did:peer:{self.me.did}',
            did_doc=did_doc_extra
        )
        if success:
            return True, p2p.their.did_doc
        else:
            return False, None



import contextlib
from typing import Union, List, Dict

import sirius_sdk
from sirius_sdk.abstract.p2p import Pairwise


@contextlib.asynccontextmanager
async def as_caller(
        called: Pairwise, thid: str, cast: Union[List, Dict], pthid: str = None, co_binding_id: str = None,
        time_to_live: int = None, logger=None, *args, **kwargs
):
    """Invoke co-protocol as Caller
        see details:
          - Feature: https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol
          - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols

    Sample code:
        async with as_caller(p2p, thid='payment-thread-id', cast=[{'role': 'payer', 'id': 'did:peer:abc123' }]) as payment_protocol:
            # check if payment protocol supports out currencies
            if "USDT" in payment_protocol.context['currencies]:
                await protocol.input(data=[{"amount": 10, "currency": "USDT"}])
                success, data, extra = await payment_protocol.wait_output()
                if success is True:
                    print('Payment was successfully terminated')

                    # Go to next steps
                    .....
                else:
                    print('Error: ' + payment_protocol.problem_report.explain)
            else:
                await payment_protocol.raise_problem(problem_code='-1', explain='Unsupported currency')
                await payment_protocol.detach()

    :param called: (required) is the called entity getting input and giving output
    :param thid: (required) protocol thread-id
    :param cast: (required) init binding data
    :param co_binding_id: (optional) extended binding id
    :param pthid: (optional) parent protocol thread-id
    :param time_to_live: (optional) max time protocol will be alive: will be automatically detached when timeout occured
    :param logger: (optional) self-designed logger
    """

    protocol = sirius_sdk.aries_rfc.Caller(
        called=called, thid=thid, pthid=pthid, time_to_live=time_to_live, logger=logger, *args, **kwargs
    )
    success, ctx = await protocol.bind(cast=cast, co_binding_id=co_binding_id)
    if not success:
        if protocol.problem_report:
            raise RuntimeError(f'Error while binding co-protocol: "{protocol.problem_report.explain}"')
        else:
            raise RuntimeError(f'Error while binding co-protocol')
    try:
        yield protocol
    finally:
        await protocol.detach()


@contextlib.asynccontextmanager
async def as_called(caller: Pairwise, request: sirius_sdk.aries_rfc.CoProtocolBind, logger=None, *args, **kwargs):
    """Invoke co-protocol as Called entity
        see details:
          - Feature: https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol
          - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols

        Sample code:
            async with as_called(p2p, request) as payment_protocol:
                # declare what currencies we supports
                await payment_protocol.attach(currencies=['USDT', 'ETH', 'BTC'])
                success, data, extra = await payment_protocol.wait_input()
                if success:
                    # Make some steps to process payment
                    .....
                    await payment_protocol.output(data={'success': True, 'report': '...'})
                else:
                    # print problem_report
                    if payment_protocol.problem_report:
                        print('Error: ' + payment_protocol.problem_report.explain)

        :param caller: (required) is the called entity getting input and giving output
        :param request: (required) caller bind request message
        :param logger: (optional) self-designed logger
    """
    protocol = await sirius_sdk.aries_rfc.Called.open(caller=caller, request=request, logger=logger, *args, **kwargs)
    try:
        yield protocol
    finally:
        await protocol.done()



"""We need a standard way for one protocol to invoke another, giving it input, getting its output, detaching, and debugging.
  details:
    - Feature: https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol
    - Concept: https://github.com/hyperledger/aries-rfcs/blob/main/concepts/0478-coprotocols
"""
from enum import Enum
from typing import Optional, Union, List, Dict, Any, Tuple

import sirius_sdk
from sirius_sdk.base import AbstractStateMachine
from sirius_sdk.hub.coprotocols import CoProtocolThreadedP2P
from sirius_sdk.abstract.p2p import Pairwise
from sirius_sdk.errors.exceptions import SiriusTimeoutIO

from .messages import CoProtocolBind, CoProtocolAttach, CoProtocolInput, CoProtocolOutput, \
    CoProtocolDetach, CoProtocolProblemReport


class AttachContext(dict):
    pass


class Caller(AbstractStateMachine):
    """The caller role is played by the entity giving input and getting output.
    The called is the entity getting input and giving output.

    see details:
      - https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol
    """

    class State(Enum):
        # https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol#states
        NULL = 'NULL'
        DETACHED = 'DETACHED'
        ATTACHED = 'ATTACHED'
        DONE = 'DONE'

    def __init__(
            self, called: Pairwise, thid: str, pthid: str = None,
            time_to_live: int = None, logger=None, *args, **kwargs
    ):
        """Initialize Caller state-machine environment

        :param called (required): is the entity getting input and giving output
        :param thid (required): co-protocol thread-id
        :param pthid (optional): parent co-protocol thread-id
        :param time_to_live (optional): time the state-machine is alive,
        :param logger (optional): state-machine logger
        """
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self._problem_report = None
        self.__time_to_live = time_to_live
        self.__called = called
        self.__thid = thid
        self.__pthid = pthid
        self.__state = self.State.NULL
        self.__context: Optional[AttachContext] = None
        self._coprotocol: Optional[CoProtocolThreadedP2P] = None
        self._coprotocol_child: Optional[CoProtocolThreadedP2P] = None
        self._problem_report: Optional[CoProtocolProblemReport] = None

    @property
    def state(self) -> State:
        """Current State"""
        return self.__state

    @property
    def context(self) -> Optional[AttachContext]:
        """Context that was retrieved from [Called] entity after bind"""
        return self.__context

    @property
    def thid(self) -> str:
        """Thread-ID

          see details:
            - https://github.com/hyperledger/aries-rfcs/tree/main/concepts/0008-message-id-and-threading#threaded-messages
        """
        return self.__thid

    @property
    def pthid(self) -> str:
        """Parent Thread-ID

          see details:
            - https://github.com/hyperledger/aries-rfcs/tree/main/concepts/0008-message-id-and-threading#threaded-messages
            - https://github.com/hyperledger/aries-rfcs/tree/main/concepts/0478-coprotocols#coroutines
        """
        return self.__pthid

    @property
    def problem_report(self) -> CoProtocolProblemReport:
        return self._problem_report

    async def bind(
            self, cast: Union[List, Dict] = None, co_binding_id: str = None, **extra_fields
    ) -> (bool, Optional[AttachContext]):
        """The protocol begins with a bind message sent from caller to called.
          This message basically says, "I would like to interact with a new coprotocol instance

        :param cast: (optional) declaring roles and binding context
        :param co_binding_id: (optional) internal [caller-to-called] binding-id
        :param extra_fields: additional bind message fields
        :return: [operation status, context called entity was set]

        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol#messages
        """
        if self.state != self.State.ATTACHED:
            await self.log(comment='create co-protocol')
            await self._create_coprotocol()
            request = self._build_bind_request(cast, co_binding_id, **extra_fields)
            await self.log(comment='sending bind request', message=request)
            await self._new_state(self.State.DETACHED)
            await self._coprotocol.send(request)
            try:
                while True:
                    response, _, _ = await self._coprotocol.get_one()
                    await self.log(comment='received attach response', message=response)
                    if isinstance(response, CoProtocolAttach):
                        self.__context = AttachContext()
                        for fld, value in response.items():
                            if not (fld.startswith('@') or fld.startswith('~')):
                                self.__context[fld] = value
                        thid = response.thid or response.id
                        await self.log(comment=f'create child co-protocol with thread-id: "thid"')
                        await self._create_child_coprotocol(thid=thid)
                        await self._new_state(self.State.ATTACHED)
                        await self._clean_coprotocol()
                        return True, self.__context
                    elif isinstance(response, CoProtocolProblemReport):
                        self._problem_report = response
                        return False, None
            except SiriusTimeoutIO:
                await self.log(comment='IO timeout occurred')
                return False, None
        else:
            await self.log(comment='already in attached state: nothing to-do')
            return True, None

    async def input(self, data: Any, **extra):
        """The caller can now kick off/invoke the protocol with an input message.

           This allows the caller to invoke the bound coprotocol instance, and to pass it any number of named inputs.
        """

        if self.state == self.State.ATTACHED:
            msg = self._build_input_message(data, **extra)
            msg.thid = self._coprotocol_child.thid
            await self.log(comment='sending input message', message=msg)
            await self._coprotocol_child.send(msg)
        else:
            await self.log(comment='state error')
            raise RuntimeError('CoProtocol must be attached first! Call bind before operate...')

    async def raise_problem(self, problem_code: str, explain: str):
        if self.state == self.State.ATTACHED:
            msg = CoProtocolProblemReport(problem_code=problem_code, explain=explain, thread_id=self.thid)
            msg.thid = self._coprotocol_child.thid
            await self.log(comment='raising problem-report', message=msg)
            await self._coprotocol_child.send(msg)
        else:
            await self.log(comment='state error')
            raise RuntimeError('CoProtocol must be attached first! Call bind before operate...')

    async def wait_output(self) -> Tuple[bool, Any, Optional[Dict]]:
        """Later, when the coprotocol instance wants to emit an output from called to caller, it uses an output message

        :return: [success, Data, extra_fields]
        """
        if self.state == self.State.ATTACHED:
            while True:
                msg, _, _ = await self._coprotocol_child.get_one()
                if isinstance(msg, CoProtocolOutput):
                    await self.log(comment='received output', message=msg)
                    data = msg.pop('data', None)
                    extra_fields = dict()
                    for fld, value in msg.items():
                        if not (fld.startswith('@') or fld.startswith('~')):
                            extra_fields[fld] = value
                    return True, data, extra_fields
                elif isinstance(msg, CoProtocolProblemReport):
                    await self.log(comment='received problem-report', message=msg)
                    self._problem_report = msg
                    return False, None, None
        else:
            await self.log(comment='state error')
            raise RuntimeError('CoProtocol must be attached first! Call bind before operate...')

    async def detach(self):
        """If a caller wants to detach, it uses a detach message.
           This leaves the coprotocol running on called; all inputs that it emits are sent to the bitbucket,
           and it advances on its normal state trajectory as if it were a wholly independent protocol:
        """
        if self.state == self.State.ATTACHED:
            await self.log(comment='detaching...')
            request = self._build_detach_request()
            request.thid = self._coprotocol_child.thid
            await self._coprotocol_child.send(request)
            await self._new_state(self.State.DETACHED)
            self.__context = None
            await self._clean_coprotocol()
            await self._clean_coprotocol_child()
            await self.log(comment='detached!')

    async def done(self):
        if self.state == self.State.ATTACHED:
            await self.log(comment='done protocol')
            await self._new_state(self.State.DONE)
            await self._clean_coprotocol()
            await self._clean_coprotocol_child()

    async def _new_state(self, new_state: State):
        """State transition"""
        if new_state != self.state:
            valid_states = []
            ignore_states = []
            if self.state == self.State.NULL:
                valid_states = [self.State.DETACHED]
                ignore_states = [self.State.DONE]
            elif self.state == self.State.DETACHED:
                valid_states = [self.State.ATTACHED]
                ignore_states = [self.State.DONE]
            elif self.state == self.State.ATTACHED:
                valid_states = [self.State.DONE, self.State.DETACHED]
                ignore_states = [self.State.DONE]
            elif self.state == self.State.DONE:
                valid_states = [self.State.DETACHED]
                ignore_states = []
            if new_state in valid_states:
                await self.log(comment=f'Change status from "{self.__state.value}" to "{new_state.value}"')
                self.__state = new_state
            elif new_state in ignore_states:
                return
            else:
                raise RuntimeError(f'Invalid state transition from "{self.__state.value}" to "{new_state.value}"')

    def _build_bind_request(
            self, cast: Union[List, Dict] = None, co_binding_id: str = None, **extra_fields
    ) -> CoProtocolBind:
        request = CoProtocolBind(
            id_=self.__thid, thid=self.__thid, pthid=self.__pthid, cast=cast, co_binding_id=co_binding_id
        )
        if extra_fields:
            for fld, value in extra_fields.items():
                request[fld] = value
        return request

    def _build_input_message(self, data: Any, **extra) -> CoProtocolInput:
        msg = CoProtocolInput(pthid=self.__thid)
        msg['data'] = data
        for fld, value in extra.items():
            msg[fld] = value
        return msg

    def _build_detach_request(self) -> CoProtocolDetach:
        request = CoProtocolDetach(pthid=self.__thid)
        return request

    async def _create_coprotocol(self) -> CoProtocolThreadedP2P:
        self._coprotocol = CoProtocolThreadedP2P(
            thid=self.__thid,
            to=self.__called,
            pthid=self.__pthid,
            time_to_live=self.time_to_live
        )
        self._register_for_aborting(self._coprotocol)
        return self._coprotocol

    async def _create_child_coprotocol(self, thid: str) -> CoProtocolThreadedP2P:
        self._coprotocol_child = CoProtocolThreadedP2P(
            thid=thid,
            to=self.__called,
            pthid=self.__thid,
            time_to_live=self.time_to_live
        )
        self._register_for_aborting(self._coprotocol_child)
        return self._coprotocol_child

    async def _clean_coprotocol(self):
        if self._coprotocol is not None:
            try:
                await self._coprotocol.clean()
                self._unregister_for_aborting(self._coprotocol)
            finally:
                self._coprotocol = None

    async def _clean_coprotocol_child(self):
        if self._coprotocol_child is not None:
            try:
                await self._coprotocol_child.clean()
                self._unregister_for_aborting(self._coprotocol_child)
            finally:
                self._coprotocol_child = None


class Called(AbstractStateMachine):

    class State(Enum):
        # https://github.com/hyperledger/aries-rfcs/tree/main/features/0482-coprotocol-protocol#states
        NULL = 'NULL'
        ATTACHED = 'ATTACHED'
        DONE = 'DONE'

    class CoProtocolDetachedByCaller(RuntimeError):
        pass

    def __init__(self, *args, **kwargs):
        """Initialize Called state-machine environment

          Do nat create State-Machine with new constructor,
          !! Use OPEN method instead !!
        """
        super().__init__(*args, **kwargs)
        self._problem_report = None
        self._coprotocol: Optional[CoProtocolThreadedP2P] = None
        self.__caller: Optional[Pairwise] = None
        self.__bind_context = None
        self.__context: Optional[AttachContext] = None
        self.__state = self.State.NULL

    @property
    def caller(self) -> Pairwise:
        return self.__caller

    @property
    def context(self) -> Optional[AttachContext]:
        return self.__context

    @property
    def state(self) -> State:
        return self.__state

    @property
    def problem_report(self) -> CoProtocolProblemReport:
        return self._problem_report

    @staticmethod
    async def open(
            caller: Pairwise, request: CoProtocolBind,
            logger=None, *args, **kwargs
    ) -> "Called":
        """The [Called] is the entity getting input and giving output.

        :param caller: Entity who invoke co-protocol
        :param request: initialize message
        :param logger: external logger
        :return: state-machine instance
        """

        inst = Called(
            time_to_live=None,  # called protocol is terminated with caller entity (detach),
            logger=logger,
            *args, **kwargs
        )
        inst.__caller = caller
        inst.__context = AttachContext()
        for fld, value in request.items():
            if not (fld.startswith('@') or fld.startswith('~')):
                inst.__context[fld] = value
        await inst.log(comment='open', message=request)
        await inst._create_coprotocol(thid=request.thid or request.id)
        return inst

    async def attach(self, **fields):
        msg = self._build_attach_message(**fields)
        await self.log(comment='attach', message=msg)
        await sirius_sdk.send_to(msg, to=self.__caller)
        self.__state = self.State.ATTACHED

    async def done(self):
        await self.log(comment='done')
        await self._clean_coprotocol()
        self.__state = self.State.DONE

    async def wait_input(self) -> (bool, Any, Optional[Dict]):
        """Later, when the coprotocol instance wants to emit an output from called to caller, it uses an output message

        :return: [success, Data, extra_fields]
        :raises CoProtocolDetachedByCaller if Caller was called detach
        """
        if self.state == self.State.ATTACHED:
            while True:
                msg, _, _ = await self._coprotocol.get_one()
                if isinstance(msg, CoProtocolInput):
                    await self.log(comment='received input', message=msg)
                    data = msg.pop('data', None)
                    extra_fields = dict()
                    for fld, value in msg.items():
                        if not (fld.startswith('@') or fld.startswith('~')):
                            extra_fields[fld] = value
                    return True, data, extra_fields
                elif isinstance(msg, CoProtocolProblemReport):
                    await self.log(comment='received problem-report', message=msg)
                    self._problem_report = msg
                    return False, None, None
                elif isinstance(msg, CoProtocolDetach):
                    await self.done()
                    raise self.CoProtocolDetachedByCaller('Called entity received DETACH message from Caller')
        else:
            await self.log(comment='state error')
            raise RuntimeError('CoProtocol must be attached first! Call bind before operate...')

    async def output(self, data: Any, **extra):
        """The called can raise output message.
        """

        if self.state == self.State.ATTACHED:
            msg = self._build_output_message(data, **extra)
            msg.thid = self._coprotocol.thid
            await self.log(comment='sending output message', message=msg)
            await self._coprotocol.send(msg)
        else:
            await self.log(comment='state error')
            raise RuntimeError('CoProtocol must be attached first! Call bind before operate...')

    async def raise_problem(self, problem_code: str, explain: str):
        if self.state == self.State.ATTACHED:
            msg = CoProtocolProblemReport(problem_code=problem_code, explain=explain, thread_id=self._coprotocol.thid)
            await self._coprotocol.send(msg)
        else:
            await self.log(comment='state error')
            raise RuntimeError('CoProtocol must be attached first! Call bind before operate...')

    def _build_attach_message(self, **fields) -> CoProtocolAttach:
        msg = CoProtocolAttach(thid=self._coprotocol.thid)
        for fld, value in fields.items():
            msg[fld] = value
        return msg

    def _build_output_message(self, data: Any, **extra) -> CoProtocolOutput:
        msg = CoProtocolOutput(thid=self._coprotocol.thid)
        msg['data'] = data
        for fld, value in extra.items():
            msg[fld] = value
        return msg

    async def _create_coprotocol(self, thid: str) -> CoProtocolThreadedP2P:
        self._coprotocol = CoProtocolThreadedP2P(
            thid=thid,
            to=self.__caller,
            time_to_live=self.time_to_live
        )
        self._register_for_aborting(self._coprotocol)
        return self._coprotocol

    async def _clean_coprotocol(self):
        if self._coprotocol is not None:
            try:
                await self._coprotocol.clean()
                self._unregister_for_aborting(self._coprotocol)
            finally:
                self._coprotocol = None



from .documents import Document, EncryptedDocument
from .streams import AbstractReadOnlyStream, AbstractWriteOnlyStream, BaseStreamEncryption, StreamEncryption, \
  StreamDecryption, ReadOnlyStreamDecodingWrapper, WriteOnlyStreamEncodingWrapper

from .encoding import ConfidentialStorageEncType, JWE, KeyPair
from .errors import BaseConfidentialStorageError, StreamEOF, EncryptionError, StreamInitializationError, StreamSeekableError, \
    StreamFormatError, ConfidentialStorageTimeoutOccurred
from .components import ConfidentialStorageAuthProvider, EncryptedDataVault, ConfidentialStorageRawByteStorage, \
    VaultConfig, ConfidentialStorageRawByteStorage, StructuredDocument, DataVaultStreamWrapper, DocumentMeta, StreamMeta
from .messages import StructuredDocumentAttach, DataVaultQueryList, DataVaultResponseList, BaseConfidentialStorageMessage
from .documents import Document, EncryptedDocument
from .impl.file_system import FileSystemReadOnlyStream, FileSystemWriteOnlyStream, FileSystemRawByteStorage
from .state_machines import CalledReadOnlyStreamProtocol, CallerReadOnlyStreamProtocol, \
    CallerWriteOnlyStreamProtocol, CalledWriteOnlyStreamProtocol, CallerEncryptedDataVault, CalledEncryptedDataVault

__all__ = [
    "BaseStreamEncryption", "StreamEncryption", "StreamDecryption", "AbstractReadOnlyStream",
    "AbstractWriteOnlyStream", "CalledReadOnlyStreamProtocol", "CallerReadOnlyStreamProtocol",
    "CallerWriteOnlyStreamProtocol", "CalledWriteOnlyStreamProtocol", "Document", "EncryptedDocument",
    "BaseConfidentialStorageError", "StreamEOF", "EncryptionError", "StreamInitializationError",
    "StreamSeekableError", "StreamFormatError", "ConfidentialStorageTimeoutOccurred", "ConfidentialStorageEncType",
    "FileSystemReadOnlyStream", "FileSystemWriteOnlyStream", "DataVaultStreamWrapper", "BaseConfidentialStorageMessage",
    "ConfidentialStorageAuthProvider", "EncryptedDataVault", "FileSystemRawByteStorage",
    "Document", "EncryptedDocument", "VaultConfig", "ConfidentialStorageRawByteStorage", "StructuredDocument",
    "StructuredDocumentAttach", "ReadOnlyStreamDecodingWrapper", "WriteOnlyStreamEncodingWrapper",
    "JWE", "KeyPair", "CallerEncryptedDataVault", "CalledEncryptedDataVault", "DataVaultQueryList",
    "DataVaultResponseList", "DocumentMeta", "StreamMeta"
]



from sirius_sdk.messaging import check_for_attributes


class DIDDoc(dict):
    DID = 'did'
    DID_DOC = 'did_doc'
    VCX_DID = 'DID'
    VCX_DID_DOC = 'DIDDoc'

    def validate(self):
        check_for_attributes(
            self,
            [
                '@context',
                'publicKey',
                'service'
            ]
        )

        for publicKeyBlock in self['publicKey']:
            check_for_attributes(
                publicKeyBlock,
                [
                    'id',
                    'type',
                    'controller',
                    'publicKeyBase58'
                ]
            )

        for serviceBlock in self['service']:
            check_for_attributes(
                serviceBlock,
                [
                    'type',
                    'recipientKeys',
                    'serviceEndpoint'
                ]
            )

    def extract_service(self, high_priority: bool = True, type_: str = 'IndyAgent'):
        services = self.get("service", [])
        if services:
            ret = None
            for service in services:
                if service['type'] != type_:
                    continue
                if ret is None:
                    ret = service
                else:
                    if high_priority:
                        if service.get("priority", 0) < ret.get("priority", 0):
                            ret = service
            return ret
        else:
            return None



from enum import Enum
from typing import Optional, Union

from sirius_sdk.messaging import check_for_attributes
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage
from sirius_sdk.agent.aries_rfc.decorators import *


class Status(Enum):
    # outcome has occurred, and it was positive
    OK = 'OK'

    # no outcome is yet known
    PENDING = 'PENDING'

    # outcome has occurred, and it was negative
    FAIL = 'FAIL'


class Ack(AriesProtocolMessage, metaclass=RegisterMessage):

    PROTOCOL = 'notification'
    NAME = 'ack'

    def __init__(self, thread_id: str = None, status: Optional[Union[Status, str]] = None, *args, **kwargs):
        super(Ack, self).__init__(*args, **kwargs)
        if status is not None:
            if isinstance(status, Status):
                self['status'] = status.value
            else:
                self['status'] = status
        if thread_id is not None:
            thread = self.get(THREAD_DECORATOR, {})
            thread['thid'] = thread_id
            self[THREAD_DECORATOR] = thread

    def validate(self):
        super().validate()
        check_for_attributes(self, [THREAD_DECORATOR])
        check_for_attributes(self[THREAD_DECORATOR], ['thid'])

    @property
    def status(self) -> Optional[Status]:
        status = self.get('status', None)
        if status is None:
            return Status.OK
        elif status == Status.OK.value:
            return Status.OK
        elif status == Status.PENDING.value:
            return Status.PENDING
        elif status == Status.FAIL.value:
            return Status.FAIL
        else:
            return None



import threading
from typing import Any, List, Optional, Dict

from sirius_sdk.abstract.storage import AbstractImmutableCollection, AbstractKeyValueStorage
from sirius_sdk.hub.context import get as context_get, set as context_set


def _get_current_hub_id() -> str:
    hub = context_get('hub')
    if hub:
        return hub.global_id
    else:
        return '*'


class InMemoryImmutableCollection(AbstractImmutableCollection):

    __lock_singleton = threading.Lock()
    __selected_db_thread_safe = threading.local()
    __database_singleton = {}

    async def select_db(self, db_name: str):
        mangled_db_name = f'{_get_current_hub_id()}/{db_name}'
        self.__acquire()
        try:
            if mangled_db_name not in self.__database_singleton:
                self.__database_singleton[mangled_db_name] = []
            self.__set_current_db(mangled_db_name)
        finally:
            self.__release()

    async def add(self, value: Any, tags: dict):
        cur_db_name = self.__get_current_db()
        if cur_db_name is None:
            raise RuntimeError('Select working database at first!')
        else:
            self.__acquire()
            try:
                item = (value, tags)
                selected_db = self.__database_singleton[cur_db_name]
                selected_db.append(item)
            finally:
                self.__release()

    async def fetch(self, tags: dict, limit: int = None) -> (List[Any], int):
        cur_db_name = self.__get_current_db()
        if cur_db_name is None:
            raise RuntimeError('Select working database at first!')
        else:
            self.__acquire()
            try:
                result = []
                selected_db = self.__database_singleton[cur_db_name]
                for item in selected_db:
                    value_ = item[0]
                    tags_ = item[1]
                    if tags.items() <= tags_.items():
                        result.append(value_)
                return result
            finally:
                self.__release()

    @classmethod
    def __acquire(cls):
        cls.__lock_singleton.acquire(blocking=True)

    @classmethod
    def __release(cls):
        cls.__lock_singleton.release()

    @classmethod
    def __set_current_db(cls, db_name: str):
        cls.__selected_db_thread_safe.value = db_name

    @classmethod
    def __get_current_db(cls) -> Optional[str]:
        try:
            db = cls.__selected_db_thread_safe.value
        except AttributeError:
            return None
        else:
            return db


class InMemoryKeyValueStorage(AbstractKeyValueStorage):

    __lock_singleton = threading.Lock()
    __selected_db_context_key = 'inmemory.storage.db'
    __database_singleton = {}

    async def select_db(self, db_name: str):
        mangled_db_name = f'{_get_current_hub_id()}/{db_name}'
        self.__acquire()
        try:
            if mangled_db_name not in self.__database_singleton:
                self.__database_singleton[mangled_db_name] = {}
            self.__set_current_db(mangled_db_name)
        finally:
            self.__release()

    async def set(self, key: str, value: Any):
        cur_db_name = self.__get_current_db()
        if cur_db_name is None:
            raise RuntimeError('Select working database at first!')
        else:
            self.__acquire()
            try:
                selected_db = self.__database_singleton[cur_db_name]
                selected_db[key] = value
            finally:
                self.__release()

    async def get(self, key: str) -> Optional[Any]:
        cur_db_name = self.__get_current_db()
        if cur_db_name is None:
            raise RuntimeError('Select working database at first!')
        else:
            self.__acquire()
            try:
                selected_db = self.__database_singleton[cur_db_name]
                return selected_db.get(key, None)
            finally:
                self.__release()

    async def delete(self, key: str):
        cur_db_name = self.__get_current_db()
        if cur_db_name is None:
            raise RuntimeError('Select working database at first!')
        else:
            self.__acquire()
            try:
                selected_db = self.__database_singleton[cur_db_name]
                if key in selected_db:
                    del selected_db[key]
            finally:
                self.__release()

    async def items(self) -> Dict:
        cur_db_name = self.__get_current_db()
        if cur_db_name is None:
            raise RuntimeError('Select working database at first!')
        else:
            self.__acquire()
            try:
                selected_db = self.__database_singleton[cur_db_name]
                copied = dict(**selected_db)
                return copied
            finally:
                self.__release()

    @classmethod
    def __acquire(cls):
        cls.__lock_singleton.acquire(blocking=True)

    @classmethod
    def __release(cls):
        cls.__lock_singleton.release()

    @classmethod
    def __set_current_db(cls, db_name: str):
        context_set(cls.__selected_db_context_key, db_name)

    @classmethod
    def __get_current_db(cls) -> Optional[str]:
        try:
            db = context_get(cls.__selected_db_context_key)
        except AttributeError:
            return None
        else:
            return db



import asyncio
import uuid

import pytest

import sirius_sdk
from sirius_sdk import Agent, TheirEndpoint, Pairwise
import sirius_sdk.hub.coprotocols
from sirius_sdk.errors.exceptions import *
from tests.conftest import get_pairwise
from sirius_sdk.messaging import Message
from tests.helpers import run_coroutines, LocalCryptoManager, LocalDIDManager
from sirius_sdk.hub.defaults.default_crypto import DefaultCryptoService

from .helpers import MSG_LOG, check_msg_log, TEST_MSG_TYPES


async def routine1(
        co: sirius_sdk.hub.coprotocols.AbstractP2PCoProtocol,
        cfg: sirius_sdk.Config, **kwargs
):
    async with sirius_sdk.context(cfg):
        try:
            first_req = Message({
                '@type': TEST_MSG_TYPES[0],
                'content': 'Request1'
            })
            MSG_LOG.append(first_req)
            print('#2')
            ok, resp1 = await co.switch(message=first_req)
            print('#2')
            assert ok is True
            MSG_LOG.append(resp1)
            ok, resp2 = await co.switch(
                message=Message({
                    '@type': TEST_MSG_TYPES[2],
                    'content': 'Request2'
                })
            )
            assert ok is True
            MSG_LOG.append(resp2)
        except Exception as e:
            raise


async def routine2(
        co: sirius_sdk.hub.coprotocols.AbstractP2PCoProtocol,
        cfg: sirius_sdk.Config, **kwargs
):
    async with sirius_sdk.context(cfg):
        try:
            await asyncio.sleep(1)
            print('#1')
            ok, resp1 = await co.switch(
                message=Message({
                    '@type': TEST_MSG_TYPES[1],
                    'content': 'Response1'
                })
            )
            print('#2')
            assert ok is True
            MSG_LOG.append(resp1)
            await co.send(
                message=Message({
                    '@type': TEST_MSG_TYPES[3],
                    'content': 'End'
                })
            )
            print('')
        except Exception as e:
            raise


@pytest.mark.skip
@pytest.mark.asyncio
async def test__their_endpoint_coprotocol(mediator_uri: str, mediator_verkey: str):

    crypto1 = LocalCryptoManager()
    crypto2 = LocalCryptoManager()
    my_vk1 = await crypto1.create_key()
    my_vk2 = await crypto2.create_key()
    cfg1 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk1, mediator_verkey).\
        override_crypto(dependency=crypto1).override_did(dependency=LocalDIDManager())
    cfg2 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk2, mediator_verkey).\
        override_crypto(dependency=crypto2).override_did(dependency=LocalDIDManager())
    print('#')

    async with sirius_sdk.context(cfg1):
        agent1_endpoint = [e for e in await sirius_sdk.endpoints() if e.is_default][0].address
    print('')
    async with sirius_sdk.context(cfg2):
        agent2_endpoint = [e for e in await sirius_sdk.endpoints() if e.is_default][0].address

    # FIRE!!!
    their1 = TheirEndpoint(agent2_endpoint, my_vk2)
    their2 = TheirEndpoint(agent1_endpoint, my_vk1)
    co1 = sirius_sdk.hub.coprotocols.CoProtocolP2PAnon(my_vk1, their1, ['test_protocol'])
    co2 = sirius_sdk.hub.coprotocols.CoProtocolP2PAnon(my_vk2, their2, ['test_protocol'])
    MSG_LOG.clear()
    await run_coroutines(
        routine1(co1, cfg1),
        routine2(co2, cfg2),
        timeout=15
    )
    check_msg_log()


@pytest.mark.skip
@pytest.mark.asyncio
async def test__threadbased_coprotocol(mediator_uri: str, mediator_verkey: str):
    crypto1 = LocalCryptoManager()
    crypto2 = LocalCryptoManager()
    my_vk1 = await crypto1.create_key()
    my_vk2 = await crypto2.create_key()
    cfg1 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk1, mediator_verkey). \
        override_crypto(dependency=crypto1).override_did(dependency=LocalDIDManager(crypto1))
    cfg2 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk2, mediator_verkey). \
        override_crypto(dependency=crypto2).override_did(dependency=LocalDIDManager(crypto2))

    async with sirius_sdk.context(cfg1):
        # Get endpoints
        agent1_endpoint = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await sirius_sdk.DID.create_and_store_my_did()

    async with sirius_sdk.context(cfg2):
        # Get endpoints
        agent2_endpoint = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        # Init pairwise list #1
        did2, verkey2 = await sirius_sdk.DID.create_and_store_my_did()

    async with sirius_sdk.context(cfg1):
        await sirius_sdk.DID.store_their_did(did2, verkey2)
    async with sirius_sdk.context(cfg2):
        await sirius_sdk.DID.store_their_did(did1, verkey1)

    # Init pairwise list #2
    pairwise1 = Pairwise(
        me=Pairwise.Me(
            did=did1,
            verkey=verkey1
        ),
        their=Pairwise.Their(
            did=did2,
            label='Label-2',
            endpoint=agent2_endpoint,
            verkey=verkey2
        )
    )
    pairwise2 = Pairwise(
        me=Pairwise.Me(
            did=did2,
            verkey=verkey2
        ),
        their=Pairwise.Their(
            did=did1,
            label='Label-1',
            endpoint=agent1_endpoint,
            verkey=verkey1
        )
    )

    thread_id = 'threadbased_coprotocol-' + uuid.uuid4().hex
    co1 = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thread_id, pairwise1)
    co2 = sirius_sdk.hub.coprotocols.CoProtocolThreadedP2P(thread_id, pairwise2)
    MSG_LOG.clear()
    await run_coroutines(
        routine1(co1, cfg1),
        routine2(co2, cfg2),
        timeout=15
    )
    check_msg_log()



import asyncio
import queue
import threading
from typing import Optional, Coroutine


class BackgroundScheduler:

    __thread_singleton: Optional[threading.Thread] = None
    __thread_queue: Optional[queue.Queue] = None
    __thread_event_loop: Optional[asyncio.AbstractEventLoop] = None
    __lock_singleton = threading.Lock()

    @classmethod
    def schedule(cls, coro: Coroutine, task_hash: str = None):
        event_loop = cls.__get_event_loop()
        coro_broker = cls.__schedule_coroutine(coro, task_hash)
        asyncio.run_coroutine_threadsafe(coro_broker, loop=event_loop)

    @classmethod
    def unschedule(cls, task_hash: str):
        event_loop = cls.__get_event_loop()
        coro_broker = cls.__unschedule_coroutine(task_hash)
        asyncio.run_coroutine_threadsafe(coro_broker, loop=event_loop)

    @classmethod
    def __thread_routine(cls):
        asyncio.set_event_loop(cls.__thread_event_loop)
        asyncio.get_event_loop().run_forever()

    @classmethod
    def __get_event_loop(cls) -> asyncio.AbstractEventLoop:
        cls.__lock_singleton.acquire(blocking=True)
        try:
            if cls.__thread_queue is None:
                cls.__thread_queue = queue.Queue()
            if cls.__thread_event_loop is None:
                cls.__thread_event_loop = asyncio.new_event_loop()
            if cls.__thread_singleton is not None and not cls.__thread_singleton.is_alive():
                cls.__thread_singleton = None
            if cls.__thread_singleton is None:
                cls.__thread_singleton = threading.Thread(target=cls.__thread_routine)
                cls.__thread_singleton.daemon = True
                cls.__thread_singleton.start()
            return cls.__thread_event_loop
        finally:
            cls.__lock_singleton.release()

    @classmethod
    async def __schedule_coroutine(cls, coro: Coroutine, task_hash: str = None):
        if task_hash is None:
            tasks = []
        else:
            tasks_with_hash = [task for task in asyncio.all_tasks(cls.__thread_event_loop) if not task.done() and hasattr(task, 'hash')]
            tasks = [task for task in tasks_with_hash if task.hash == task_hash]
        if tasks:
            tsk = tasks[0]
            tsk.counter += 1
        else:
            tsk = asyncio.create_task(coro)
            tsk.hash = task_hash
            tsk.counter = 1

    @classmethod
    async def __unschedule_coroutine(cls, task_hash: str):
        tasks_with_hash = [task for task in asyncio.all_tasks(cls.__thread_event_loop) if not task.done() and hasattr(task, 'hash')]
        tasks = [task for task in tasks_with_hash if task.hash == task_hash]
        if tasks:
            tsk = tasks[0]
            tsk.counter -= 1
            if tsk.counter < 1:
                tsk.cancel()



import json
import uuid
import asyncio

import pytest
import sirius_sdk
from sirius_sdk import Agent
from sirius_sdk.messaging import Message

from tests.helpers import ServerTestSuite
from tests.conftest import get_pairwise2


async def run_sender_with_delay(
        sdk: dict, their_vk: str, endpoint: str, my_vk: str,
        msg: sirius_sdk.aries_rfc.Message, delay: int = 3
):
    async with sirius_sdk.context(sdk['server_address'], sdk['credentials'], sdk['p2p']):
        await asyncio.sleep(delay)
        try:
            await sirius_sdk.send(
                msg, their_vk=their_vk, endpoint=endpoint,
                my_vk=my_vk, routing_keys=[]
            )
        except Exception as e:
            raise


@pytest.mark.asyncio
async def test_message_delivery(test_suite: ServerTestSuite):
    sdk_me = test_suite.get_agent_params('agent1')
    sdk_their = test_suite.get_agent_params('agent2')
    expected_content = 'Hello-' + uuid.uuid4().hex

    mediate_vk_bytes, mediate_sk_bytes = sirius_sdk.encryption.create_keypair()
    mediate_vk = sirius_sdk.encryption.bytes_to_b58(mediate_vk_bytes)
    mediate_sk = sirius_sdk.encryption.bytes_to_b58(mediate_sk_bytes)
    p2p = await get_pairwise2((sdk_me, 'agent1'), (sdk_their, 'agent2'))

    msg = sirius_sdk.aries_rfc.BasicMessage(
        content=expected_content
    )
    fut = asyncio.ensure_future(
        run_sender_with_delay(sdk_me, mediate_vk, p2p.their.endpoint, p2p.me.verkey, msg)
    )

    async with sirius_sdk.context(sdk_their['server_address'], sdk_their['credentials'], sdk_their['p2p']):
        listener = await sirius_sdk.subscribe()
        event = await listener.get_one(timeout=5)
    assert event
    fwd = Message(event['message'])
    assert fwd.protocol == 'routing'
    assert fwd.name == 'forward'
    assert fwd['to'] == mediate_vk
    s, sender_vk, recip_vk = sirius_sdk.encryption.unpack_message(fwd['msg'], mediate_vk, mediate_sk)
    msg = json.loads(s)
    assert msg['content'] == expected_content
    assert sender_vk == p2p.me.verkey
    assert recip_vk == mediate_vk



import datetime
import json
from enum import Enum
from dataclasses import dataclass
from collections import UserDict
from typing import List, Optional, Dict, Any, Union

from sirius_sdk.agent.aries_rfc import Ack
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, \
    VALID_DOC_URI, AriesProblemReport
from .documents import EncryptedDocument
from .components import StructuredDocument, HMAC, VaultConfig
from .utils import datetime_to_utc_str


class BaseConfidentialStorageMessage(AriesProtocolMessage, metaclass=RegisterMessage):
    """Aries feature 0750 Messages implementation

    """
    DOC_URI = VALID_DOC_URI[0]
    PROTOCOL = 'storage'


class ConfidentialStorageMessageProblemReport(AriesProblemReport, metaclass=RegisterMessage):
    PROTOCOL = BaseConfidentialStorageMessage.PROTOCOL


@dataclass
class StructuredDocumentAttach:
    """A structured document is used to store application data as well as metadata about the application data.
    This information is typically encrypted and then stored on the data vault.
    """

    class Meta(UserDict):
        """Key-value metadata associated with the structured document."""

        def __init(
                self, created: Union[str, datetime.datetime] = None,
                content_type: str = None, chunks: int = None, **kwargs
        ):
            super().__init__(**kwargs)
            if created is not None:
                if isinstance(created, datetime.datetime):
                    self['created'] = datetime_to_utc_str(created)
                else:
                    self['created'] = created
            if content_type is not None:
                self['contentType'] = content_type
            if chunks is not None:
                self['chunks'] = chunks

        @property
        def created(self) -> Optional[str]:
            return self.get('created', None)

        @property
        def content_type(self) -> Optional[str]:
            return self.get('contentType', None)

        @property
        def chunks(self) -> Optional[int]:
            return self.get('chunks', None)

    class Stream(UserDict):
        """The stream identifier MUST be a URI that references a stream on the same data vault.
        Once the stream has been written to the data vault, the content identifier MUST be updated such that
        it is a valid hashlink. To allow for streaming encryption, the value of the digest for the stream is assumed
        to be unknowable until after the stream has been written.
        The hashlink MUST exist as a content hash for the stream that has been written to the data vault."""

        def __init__(self, id_: str = None, **kwargs):
            super().__init__(**kwargs)
            if id_ is not None:
                self['id'] = id_

        @property
        def id(self) -> Optional[str]:
            return self.get('id', None)

    class Indexed(UserDict):

        def __init__(self, sequence: int = None, attributes: List[str] = None, hmac: Union[dict, HMAC] = None, **kwargs):
            super().__init__(**kwargs)
            if sequence is not None:
                self['sequence'] = sequence
            if attributes is not None:
                self['attributes'] = attributes
            if isinstance(hmac, HMAC):
                self['hmac'] = hmac.as_json()
            elif isinstance(hmac, dict):
                self['hmac'] = hmac

        @property
        def sequence(self) -> Optional[int]:
            return self.get('sequence', None)

        @property
        def attributes(self) -> List[str]:
            return self.get('attributes', [])

        @property
        def hmac(self) -> Optional[HMAC]:
            js = self.get('hmac', None)
            if js:
                return HMAC(id=js.get('id', None), type=js.get('type', None))
            else:
                return None

    # An identifier for the structured document.
    # The value is required and MUST be a Base58-encoded 128-bit random value.
    id: Optional[str] = None
    urn: Optional[str] = None
    # A unique counter for the data vault in order to ensure that clients are properly synchronized to the data vault.
    # The value is required and MUST be an unsigned 64-bit number.
    sequence: int = None

    # Key-value metadata associated with the structured document.
    meta: Meta = None

    # Streams can be used to store images, video, backup files, and any other binary data of arbitrary length.
    stream: Stream = None

    # Indexed metadata
    indexed: Optional[List[Indexed]] = None

    # Key-value content for the structured document.
    _content: Optional[Union[dict, str]] = None

    # A JSON Web Encryption or COSE Encrypted value that, if decoded, results in the corresponding StructuredDocument.
    _jwm: Optional[dict] = None

    _cached: Optional[EncryptedDocument] = None

    @staticmethod
    def create_from(src: StructuredDocument, sequence: int) -> "StructuredDocumentAttach":
        inst = StructuredDocumentAttach(id=src.id, sequence=sequence, urn=src.urn)
        if src.indexed:
            inst.indexed = [
                StructuredDocumentAttach.Indexed(
                    sequence=index.sequence or i,
                    attributes=index.attributes,
                    hmac=index.hmac
                )
                for i, index in enumerate(src.indexed)
            ]
        if src.meta:
            inst.meta = StructuredDocumentAttach.Meta(**src.meta)
        if src.stream:
            inst.stream = StructuredDocumentAttach.Stream(id_=src.id)
        if src.content:
            if src.content.encrypted:
                inst._jwm = src.content.jwm
            else:
                inst._content = src.content.content or b''
        return inst

    @property
    def document(self) -> Optional[EncryptedDocument]:
        if self._cached is not None:
            return self._cached
        if self._content is not None:
            doc = EncryptedDocument()
            doc.content = self._content
            self._cached = doc
            return doc
        elif self._jwm is not None:
            recipients = self._jwm.get('recipients', [])
            target_verkeys = []
            for item in recipients:
                target_verkeys.append(item['header']['kid'])
            doc = EncryptedDocument(target_verkeys=target_verkeys)
            doc.content = json.dumps(self._jwm).encode()
            doc.encrypted = True
            self._cached = doc
            return doc
        else:
            return None

    @document.setter
    def document(self, doc: EncryptedDocument):
        self._cached = None
        self._content = None
        self._jwm = None
        if doc.encrypted and isinstance(doc.content, bytes):
            self._jwm = json.loads(doc.content.decode())
        elif isinstance(doc.content, dict):
            self._content = doc.content
        elif isinstance(doc.content, str):
            self._content = {'message': doc.content}

    def as_json(self) -> dict:
        js = {'id': self.urn}
        if self.meta is not None:
            js['meta'] = dict(self.meta)
        if self.stream is not None:
            js['stream'] = dict(self.stream)
        if self.indexed is not None:
            js['indexed'] = [dict(ind) for ind in self.indexed]
        if self._content is not None:
            content_js = {'id': self.id}
            if self._content:
                if isinstance(self._content, bytes):
                    message = self._content.decode()
                elif isinstance(self._content, str) or isinstance(self._content, dict) or isinstance(self._content, list):
                    message = self._content
                else:
                    message = None
                if message is not None:
                    content_js['message'] = message
            js['content'] = content_js
        if self._jwm is not None:
            js['jwm'] = self._jwm
        return js

    def from_json(self, js: dict):
        self._cached = None
        # Id restored from sub-attrs
        self.id = js.get('content', {}).get('id') or js.get('stream', {}).get('id')
        self.urn = js.get('id', None)
        if not self.id:
            self.id = self.urn
        self.sequence = js.get('sequence', None)
        self.meta = self.Meta(**js['meta']) if 'meta' in js else None
        self.stream = self.Stream(**js['stream']) if 'stream' in js else None
        if 'indexed' in js:
            self.indexed = [self.Indexed(**kwargs) for kwargs in js['indexed']]
        else:
            self.indexed = None
        self._jwm = None
        self._content = None
        if 'jwm' in js:
            self._jwm = js['jwm'] if 'jwm' in js else None
        if 'content' in js:
            self._content = js['content'].get('message', None)


class StreamOperation(BaseConfidentialStorageMessage):

    NAME = 'stream-operation'

    class OperationCode(Enum):
        OPEN = 'open'
        CLOSE = 'close'
        SEEK_TO_CHUNK = 'seek_to_chunk'
        READ_CHUNK = 'read_chunk'
        WRITE_CHUNK = 'write_chunk'
        TRUNCATE = 'truncate'

    def __init__(self, operation: Union[str, OperationCode] = None, params=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(operation, self.OperationCode):
            self['operation'] = operation.value
        else:
            self['operation'] = operation
        if params:
            self['params'] = params

    @property
    def operation(self) -> Optional[OperationCode]:
        op = self.get('operation')
        for item in self.OperationCode:
            if op == item.value:
                return item
        return None

    @property
    def params(self) -> Optional[Dict]:
        return self.get('params', None)


class StreamOperationResult(BaseConfidentialStorageMessage):

    NAME = 'stream-operation-result'

    def __init__(self, operation: Union[str, StreamOperation.OperationCode] = None, params=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(operation, StreamOperation.OperationCode):
            self['operation'] = operation.value
        else:
            self['operation'] = operation
        if params:
            self['params'] = params

    @property
    def params(self) -> Optional[Dict]:
        return self.get('params', None)


class BaseDataVaultOperation(BaseConfidentialStorageMessage):

    def __init__(
            self, uri: str = None, meta: dict = None, attributes: dict = None,
            chunk_size: int = None, vault: str = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if uri:
            self['uri'] = uri
        if meta:
            self['meta'] = meta
        if attributes:
            self['attributes'] = attributes
        if chunk_size is not None:
            self['chunk_size'] = chunk_size
        if vault is not None:
            self['vault'] = vault

    @property
    def uri(self) -> Optional[str]:
        return self.get('uri', None)

    @property
    def meta(self) -> Optional[dict]:
        return self.get('meta', None)

    @property
    def attributes(self) -> Optional[dict]:
        return self.get('attributes', None)

    @property
    def chunk_size(self) -> Optional[int]:
        return self.get('chunk_size', None)

    @property
    def vault(self) -> Optional[str]:
        return self.get('vault', None)


class DataVaultQueryList(BaseDataVaultOperation):
    NAME = 'vault-query'


class DataVaultResponseList(BaseDataVaultOperation):
    NAME = 'vault-response'

    def __init__(self, vaults: Union[List[VaultConfig], List[dict]] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if vaults is not None and isinstance(vaults, list):
            collection = []
            for cfg in vaults:
                if isinstance(cfg, VaultConfig):
                    collection.append(cfg.as_json())
                elif isinstance(cfg, dict):
                    collection.append(cfg)
            self['vaults'] = collection

    @property
    def vaults(self) -> List[VaultConfig]:
        collection = [VaultConfig.create_from_json(js) for js in self.get('vaults', [])]
        return collection


class DataVaultOpen(BaseDataVaultOperation):
    NAME = 'vault-open'


class DataVaultClose(BaseDataVaultOperation):
    NAME = 'vault-close'


class DataVaultCreateStream(BaseDataVaultOperation):
    NAME = 'vault-create-stream'


class DataVaultCreateDocument(BaseDataVaultOperation):
    NAME = 'vault-create-document'


class DataVaultUpdateResource(BaseDataVaultOperation):
    NAME = 'data-vault-update-resource'


class DataVaultLoadResource(BaseDataVaultOperation):
    NAME = 'data-vault-load-resource'


class DataVaultRemoveResource(BaseDataVaultOperation):
    NAME = 'data-vault-remove-resource'


class DataVaultList(BaseDataVaultOperation):
    NAME = 'data-vault-list'

    def __init__(self, filters: dict = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if filters is not None:
            self['filters'] = filters

    @property
    def filters(self) -> dict:
        return self.get('filters', {})


class DataVaultBindStreamForReading(BaseDataVaultOperation):
    NAME = 'data-vault-bind-stream-reading'

    def __init__(self, uri: str, co_binding_id: str = None, *args, **kwargs):
        super().__init__(uri=uri, *args, **kwargs)
        if co_binding_id is not None:
            self['co_binding_id'] = co_binding_id

    @property
    def co_binding_id(self) -> Optional[str]:
        return self.get('co_binding_id', None)


class DataVaultBindStreamForWriting(BaseDataVaultOperation):
    NAME = 'data-vault-bind-stream-writing'

    def __init__(self, uri: str, co_binding_id: str = None, *args, **kwargs):
        super().__init__(uri=uri, *args, **kwargs)
        if co_binding_id is not None:
            self['co_binding_id'] = co_binding_id

    @property
    def co_binding_id(self) -> Optional[str]:
        return self.get('co_binding_id', None)


class DataVaultSaveDocument(BaseDataVaultOperation):
    NAME = 'data-vault-save-document'

    @property
    def document(self) -> Optional[EncryptedDocument]:
        attachment = self.get('~attach', [])
        if isinstance(attachment, list):
            attachment = attachment[0]
        if 'content' in attachment:
            content = attachment['content']
            if 'message' in content:
                return EncryptedDocument(content=content['message'])
            else:
                return EncryptedDocument(content=content)
        elif 'jwm' in attachment:
            jwm = attachment['jwm']
            doc = EncryptedDocument(content=json.dumps(jwm).encode())
            doc.encrypted = True
            return doc
        else:
            return None

    @document.setter
    def document(self, value: EncryptedDocument):
        if value.encrypted:
            if isinstance(value.content, bytes):
                jwm = json.loads(value.content.decode())
            elif isinstance(value.content, str):
                jwm = json.loads(value.content)
            else:
                jwm = value.content
            self['~attach'] = {
                'jwm': jwm
            }
        else:
            if isinstance(value.content, bytes):
                content = value.content.decode()
            else:
                content = value.content
            if isinstance(content, str):
                self['~attach'] = {
                    'content': {'message': content}
                }
            else:
                self['~attach'] = {
                    'content': content
                }


class DataVaultOperationAck(Ack):
    PROTOCOL = BaseDataVaultOperation.PROTOCOL


class StructuredDocumentMessage(BaseConfidentialStorageMessage):
    NAME = 'structured-documents'

    def __init__(self, documents: List[StructuredDocumentAttach] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if documents is not None:
            attach = []
            for i, doc in enumerate(documents):
                attach.append({'@id': f'doc-{i}', 'data': doc.as_json()})
            self['doc~attach'] = attach

    @property
    def documents(self) -> List[StructuredDocumentAttach]:
        collection = []
        doc_attachments = self.get('doc~attach', {})
        if isinstance(doc_attachments, dict):
            doc_attachments = [doc_attachments]
        for attach in doc_attachments:
            doc = StructuredDocumentAttach()
            doc.from_json(attach['data'])
            collection.append(doc)
        return collection



from .messages import BusEvent, BusPublishResponse, BusBindResponse, BusPublishRequest, BusUnsubscribeRequest, \
    BusSubscribeRequest, BusProblemReport, BusOperation


__all__ = [
    "BusEvent", "BusPublishResponse", "BusBindResponse", "BusPublishRequest", "BusUnsubscribeRequest",
    "BusSubscribeRequest", "BusProblemReport", "BusOperation"
]



import datetime

import pytest

import sirius_sdk
from sirius_sdk.agent.aries_rfc.feature_0113_question_answer import Question, Answer
from sirius_sdk.recipes.question_answer import make_answer, ask_and_wait_answer

from tests.conftest import get_pairwise
from tests.helpers import ServerTestSuite, run_coroutines


@pytest.mark.asyncio
async def test_sane(agent1: sirius_sdk.Agent, agent2: sirius_sdk.Agent, test_suite: ServerTestSuite):

    requester = agent1
    responder = agent2
    await requester.open()
    await responder.open()
    try:
        req2resp = await get_pairwise(requester, responder)
    finally:
        await requester.close()
        await responder.close()

    params_req = test_suite.get_agent_params('agent1')
    params_resp = test_suite.get_agent_params('agent2')

    async def requester(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, *args, **kwargs):
        async with sirius_sdk.context(server_address, credentials, p2p):
            q = Question(
                valid_responses=['Yes', 'No'],
                question_text='Test question',
                question_detail='Question detail'
            )
            q.set_ttl(30)
            success, answer = await ask_and_wait_answer(q, req2resp)
            return success and answer.response == 'Yes'

    async def responder(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, *args, **kwargs):
        async with sirius_sdk.context(server_address, credentials, p2p):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                if isinstance(event.message, Question):
                    await make_answer('Yes', event.message, event.pairwise)
                    return True

    coro_requester = requester(**params_req)
    coro_responder = responder(**params_resp)
    print('Run state machines')
    results = await run_coroutines(coro_requester, coro_responder, timeout=60)
    print('Finish state machines')
    print(str(results))
    assert len(results) == 2
    assert all(ret is True for ret in results)


@pytest.mark.asyncio
async def test_timeout(agent1: sirius_sdk.Agent, agent2: sirius_sdk.Agent, test_suite: ServerTestSuite):

    requester = agent1
    responder = agent2
    await requester.open()
    await responder.open()
    try:
        req2resp = await get_pairwise(requester, responder)
    finally:
        await requester.close()
        await responder.close()

    params_req = test_suite.get_agent_params('agent1')
    results = None
    timeout = 10

    async def requester(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, *args, **kwargs):
        nonlocal results
        nonlocal timeout
        async with sirius_sdk.context(server_address, credentials, p2p):
            q = Question(
                valid_responses=['Yes', 'No'],
                question_text='Test question',
                question_detail='Question detail'
            )
            q.set_ttl(timeout)
            results = await ask_and_wait_answer(q, req2resp)

    coro_requester = requester(**params_req)
    print('Run state machines')
    stamp1 = datetime.datetime.now()
    await run_coroutines(coro_requester, timeout=2*timeout)
    stamp2 = datetime.datetime.now()
    print('Finish state machines')
    print(str(results))
    assert isinstance(results, tuple)
    assert len(results) == 2
    success, message = results
    assert success is False
    assert message is None

    delta = stamp2 - stamp1
    secs = delta.total_seconds()
    assert 4/5*timeout < secs < 2*timeout



import datetime
import hashlib
import json
import os.path
import uuid
from urllib.parse import urlparse

from typing import List, Optional, Tuple, Any

import sirius_sdk
from sirius_sdk.agent.aries_rfc.feature_0750_storage import EncryptedDataVault, StructuredDocument, \
    ConfidentialStorageAuthProvider, VaultConfig, ConfidentialStorageRawByteStorage, FileSystemRawByteStorage, \
    StreamEncryption, AbstractWriteOnlyStream, AbstractReadOnlyStream, EncryptedDocument, DataVaultStreamWrapper, \
    StreamDecryption, StreamMeta, DocumentMeta
from sirius_sdk.agent.aries_rfc.feature_0750_storage.errors import *
from sirius_sdk.agent.aries_rfc.feature_0750_storage.encoding import ConfidentialStorageEncType
from sirius_sdk.agent.aries_rfc.feature_0750_storage.impl.file_system import FileSystemWriteOnlyStream


class SimpleDataVault(EncryptedDataVault, EncryptedDataVault.Indexes):

    STORAGE_TYPE = 'data_vault'

    # Meta
    META_CREATED_ATTR = 'created'
    META_CONTENT_TYPE_ATTR = 'contentType'
    META_CREATED_CHUNKS_ATTR = 'chunks'
    RESERVED_META_ATTRS = [META_CREATED_ATTR, META_CONTENT_TYPE_ATTR, META_CREATED_CHUNKS_ATTR]

    # Attributes
    ATTR_IS_STREAM = '__is_stream'
    ATTR_CHUNKS_NUM = '__chunks_num'
    ATTR_CHUNK_SIZE = '__chunk_size'
    ATTR_URN = '__urn'
    RESERVED_ATTRIBS = [ATTR_IS_STREAM, ATTR_CHUNKS_NUM, ATTR_CHUNK_SIZE, ATTR_URN]

    def __init__(self, mounted_dir: str, auth: ConfidentialStorageAuthProvider, cfg: VaultConfig = None):
        if not os.path.isdir(mounted_dir):
            raise RuntimeError(f'Directory "{mounted_dir}" does not exists')
        super().__init__(auth, cfg)
        self.__mounted_dir = os.path.join(mounted_dir, f'vault_{auth.entity.their.did}')
        if not os.path.isdir(self.__mounted_dir):
            os.mkdir(self.__mounted_dir)
        self.__storage: Optional[ConfidentialStorageRawByteStorage] = None
        self.__cached_info: Tuple[str, dict] = ('', {})
        self.__is_open: bool = False

    @property
    def mounted_dir(self) -> str:
        return self.__mounted_dir

    @property
    def is_open(self) -> bool:
        return self.__is_open

    async def open(self):
        if not self.__is_open:
            self.__storage = await self._mounted()
            self.__is_open = True

    async def close(self):
        if self.__is_open:
            if self.__storage is not None:
                self.__storage = None
            self.__is_open = False

    async def indexes(self) -> EncryptedDataVault.Indexes:
        self.auth.validate(can_read=True)
        return self

    async def create_stream(self, uri: str, meta:  Union[dict, StreamMeta] = None, chunk_size: int = None, **attributes) -> StructuredDocument:
        self.__check_is_open()
        uri = self.__normalize_uri(uri)
        await self.__create_resource(uri, is_stream=True, meta=meta, chunk_size=chunk_size, **attributes)
        info = await self.__load_resource_info(uri)
        meta = self.__extract_meta_from_info(info)
        attrib_as_dict = self.__extract_attributed_from_info(info)
        urn = info['tags'].get(self.ATTR_URN, None)
        uri = info['id']
        return StructuredDocument(
            id_=uri, meta=meta, content=None, urn=urn,
            indexed=[StructuredDocument.Index(sequence=0, attributes=list(attrib_as_dict.keys()))],
            stream=DataVaultStreamWrapper()
        )

    async def create_document(self, uri: str, meta: Union[dict, DocumentMeta] = None, **attributes) -> StructuredDocument:
        self.__check_is_open()
        uri = self.__normalize_uri(uri)
        await self.__create_resource(uri, is_stream=False, meta=meta, **attributes)
        info = await self.__load_resource_info(uri)
        meta = self.__extract_meta_from_info(info)
        attrib_as_dict = self.__extract_attributed_from_info(info)
        urn = info['tags'].get(self.ATTR_URN, None)
        uri = info['id']
        return StructuredDocument(
            id_=uri, meta=meta,
            content=EncryptedDocument(content=None),
            urn=urn,
            indexed=[StructuredDocument.Index(sequence=0, attributes=list(attrib_as_dict.keys()))]
        )

    async def remove(self, uri: str):
        info = await self.__load_resource_info(uri)
        if info:
            uri = info['id']
        await self.__remove_resource(uri, only_infos=False)

    async def update(self, uri: str, meta: Union[dict, DocumentMeta, StreamMeta] = None, **attributes):
        self.__check_is_open()
        self.auth.validate(can_update=True)
        uri = self.__normalize_uri(uri)
        info = await self.__load_resource_info(uri)
        if info is None:
            raise DataVaultResourceMissing(f'Mission resource uri: {uri}')
        stored_meta = self.__extract_meta_from_info(info)
        cleaned_meta = {k: v for k, v in stored_meta.items() if k in self.RESERVED_META_ATTRS}
        new_meta = dict(**meta)
        new_meta.update(cleaned_meta)
        tags = info['tags']
        new_tags = self.__clean_income_attributes(**attributes)
        # save old reserved attrs: rewrite new ones
        new_tags.update({k: v for k, v in tags.items() if k in self.RESERVED_ATTRIBS})
        await sirius_sdk.NonSecrets.update_wallet_record_value(type_=self._storage_type, id_=uri, value=json.dumps(new_meta))
        await sirius_sdk.NonSecrets.update_wallet_record_tags(type_=self._storage_type, id_=uri, tags=new_tags)
        self.__clean_caches()

    async def load(self, uri: str) -> StructuredDocument:
        self.__check_is_open()
        self.auth.validate(can_read=True)
        storage = await self._mounted()
        info = await self.__load_resource_info(uri)
        if info:
            uri = info['id']
        if info is None or not await storage.exists(uri):
            raise DataVaultResourceMissing(f'Mission resource uri: {uri}')
        urn = info['tags'].get(self.ATTR_URN, None)
        meta = self.__extract_meta_from_info(info)
        is_stream = info['tags'].get(self.ATTR_IS_STREAM, None) != 'no'
        chunks_num = int(info['tags'].get(self.ATTR_CHUNKS_NUM, '0'))
        storage = await self._mounted()
        attrib_as_dict = self.__extract_attributed_from_info(info)
        indexed = [StructuredDocument.Index(sequence=0, attributes=list(attrib_as_dict.keys()))]
        if is_stream:
            if self.META_CREATED_CHUNKS_ATTR not in meta:
                meta[self.META_CREATED_CHUNKS_ATTR] = int(chunks_num)
            if self.auth.can_read or self.auth.can_write:
                stream_wrapper = DataVaultStreamWrapper(
                    readable=await self.readable(uri) if self.auth.can_read else None,
                    writable=await self.writable(uri) if self.auth.can_write else None
                )
            else:
                stream_wrapper = None
            return StructuredDocument(id_=uri, meta=meta, content=None, stream=stream_wrapper, urn=urn, indexed=indexed)
        else:
            if chunks_num > 0:
                stream = await storage.readable(uri, chunks_num)
                await stream.open()
                try:
                    doc = EncryptedDocument()
                    await doc.load(stream)
                    # try detect is encrypted
                    try:
                        o = json.loads(doc.content.decode())
                        if 'protected' in o:
                            doc.encrypted = True
                    except json.JSONDecodeError:
                        pass
                    return StructuredDocument(id_=uri, meta=meta, content=doc, urn=urn, indexed=indexed)
                finally:
                    await stream.close()
            else:
                doc = EncryptedDocument()
                doc.content = None
                return StructuredDocument(id_=uri, meta=meta, content=doc, urn=urn, indexed=indexed)

    async def save_document(self, uri: str, doc: EncryptedDocument):
        self.__check_is_open()
        self.auth.validate(can_write=True)
        info = await self.__load_resource_info(uri)
        if info is None:
            raise DataVaultResourceMissing(f'Mission resource uri: {uri}')
        if info['tags'][self.ATTR_IS_STREAM] != 'no':
            raise StreamInitializationError(f'Resource uri: "{uri}" is stream, you can not operate as document with it')
        tags = info['tags']
        uri = info['id']
        storage = await self._mounted()
        stream = await storage.writeable(uri)
        await stream.open()
        try:
            await doc.save(stream)
            tags.update({self.ATTR_CHUNKS_NUM: str(stream.chunks_num)})
            await sirius_sdk.NonSecrets.update_wallet_record_tags(
                type_=self._storage_type, id_=uri, tags=tags
            )
        finally:
            await stream.close()

    async def readable(self, uri: str) -> AbstractReadOnlyStream:
        self.__check_is_open()
        self.auth.validate(can_read=True)
        info = await self.__load_resource_info(uri)
        if info is None:
            raise DataVaultResourceMissing(f'Mission resource uri: {uri}')
        if info['tags'][self.ATTR_IS_STREAM] != 'yes':
            raise StreamInitializationError(f'Resource uri: "{uri}" is not stream')
        chunks_num = int(info['tags'].get(self.ATTR_CHUNKS_NUM, '0'))
        storage = await self._mounted()
        stream = await storage.readable(uri, chunks_num)
        return stream

    async def writable(self, uri: str) -> AbstractWriteOnlyStream:
        self.__check_is_open()
        self.auth.validate(can_write=True)
        info = await self.__load_resource_info(uri)
        if info is None:
            raise DataVaultResourceMissing(f'Mission resource uri: {uri}')
        if info['tags'][self.ATTR_IS_STREAM] != 'yes':
            raise StreamInitializationError(f'Resource uri: "{uri}" is not stream')
        storage = await self._mounted()
        stream = await storage.writeable(uri)
        stored_chunk_size = info['tags'].get(self.ATTR_CHUNK_SIZE, 'null')
        if stored_chunk_size != 'null':
            stream.chunk_size = int(stored_chunk_size)
        if isinstance(stream, FileSystemWriteOnlyStream):

            async def __on_close__(uri_: str, tags_: dict, s_: AbstractWriteOnlyStream):
                # Update stream metadata on close
                tags_.update({self.ATTR_CHUNKS_NUM: str(s_.chunks_num)})
                await sirius_sdk.NonSecrets.update_wallet_record_tags(
                    type_=self._storage_type, id_=uri_, tags=tags_
                )

            stream.on_closed = __on_close__(uri, info['tags'], stream)
        return stream

    async def filter(self, **attributes) -> List[StructuredDocument]:
        self.__check_is_open()
        opts = sirius_sdk.NonSecretsRetrieveRecordOptions()
        opts.check_all()
        limit = 10000
        kwargs = dict(type_=self._storage_type, query=dict(**attributes), options=opts)
        collection, total = await sirius_sdk.NonSecrets.wallet_search(**kwargs, limit=limit)
        if total > limit:
            collection, _ = await sirius_sdk.NonSecrets.wallet_search(**kwargs, limit=total)
        ret = []
        if collection is None:
            collection = []
        missing_resources = []
        for item in collection:
            try:
                doc = await self.load(uri=item['id'])
            except DataVaultResourceMissing:
                missing_resources.append(item['id'])
            else:
                ret.append(doc)
        for uri in missing_resources:
            await self.__remove_resource(uri, only_infos=True)
        return ret

    async def _mounted(self) -> ConfidentialStorageRawByteStorage:
        if self.__storage is None:
            if self.cfg.key_agreement is None:
                encryption = None
                decryption = None
            else:
                if not self.cfg.key_agreement.type.startswith('X25519'):
                    raise EncryptionError(f'Unsupported key agreement type: "{self.cfg.key_agreement.type}"')
                storage_id = hashlib.md5(self.__mounted_dir.encode()).hexdigest()
                # Try to load encryption settings
                encryption = await self.__load_storage_encryption(storage_id)
                if encryption is None:
                    # Initialize encryption settings
                    encryption = StreamEncryption(type_=ConfidentialStorageEncType.X25519KeyAgreementKey2019)
                    if self.cfg.key_agreement.id.startswith('did:key:'):
                        target_verkey = self.cfg.key_agreement.id.split(':')[-1]
                    else:
                        target_verkey = self.cfg.key_agreement.id
                    encryption.setup(target_verkeys=[target_verkey])
                    await self.__init_storage_encryption(storage_id, encryption)
                pass
            #
            self.__storage = FileSystemRawByteStorage(encryption=encryption)
            await self.__storage.mount(self.__mounted_dir)
        return self.__storage

    async def __load_resource_info(self, uri: str) -> Optional[dict]:
        cached_uri, cached_info = self.__cached_info
        if cached_uri == uri:
            return cached_info
        opts = sirius_sdk.NonSecretsRetrieveRecordOptions()
        opts.check_all()
        try:
            rec = await sirius_sdk.NonSecrets.get_wallet_record(
                type_=self._storage_type, id_=uri, options=opts
            )
        except:
            rec = None
        if rec is None:
            recs, _ = await sirius_sdk.NonSecrets.wallet_search(
                type_=self._storage_type, query={self.ATTR_URN: uri}, options=opts, limit=1
            )
            if recs:
                rec = recs[0]
        if rec:
            self.__cached_info = (uri, rec)
        return rec

    @property
    def _storage_type(self) -> str:
        my_id = self.cfg.id or self.cfg.reference_id
        return f'{self.STORAGE_TYPE}:{my_id}'

    @staticmethod
    def __normalize_uri(uri: str):
        p = urlparse(uri)
        path = os.path.join(p.netloc, p.path)
        while path.startswith('/'):
            path = path[1:]
        return os.path.join('file:///', path)

    def __clean_income_attributes(self, **attributes) -> dict:
        return {k: v for k, v in attributes.items() if k not in self.RESERVED_ATTRIBS}

    @staticmethod
    def __extract_meta_from_info(info: dict) -> dict:
        meta = {}
        if info and 'value' in info:
            try:
                meta = json.loads(info['value'])
            except json.JSONDecodeError:
                pass
        return meta

    def __extract_attributed_from_info(self, info: dict) -> dict:
        tags = info.get('tags', {})
        return {k: v for k, v in tags.items() if k not in self.RESERVED_ATTRIBS}

    def __check_is_open(self):
        if not self.__is_open:
            raise DataVaultStateError('You should open Vault at first!')

    async def __load_storage_encryption(self, storage_id: str) -> Optional[StreamEncryption]:
        storage_type = f'{self._storage_type}/storage'
        opts = sirius_sdk.NonSecretsRetrieveRecordOptions(retrieve_value=True)
        try:
            rec = await sirius_sdk.NonSecrets.get_wallet_record(
                type_=storage_type, id_=storage_id, options=opts
            )
            js = json.loads(rec['value'])
            cek = sirius_sdk.encryption.b58_to_bytes(js['cek'])
            jwe = js['jwe']
            enc = StreamEncryption.from_jwe(jwe, cek)
            return enc
        except:
            return None

    async def __init_storage_encryption(self, storage_id: str, enc: StreamEncryption):
        storage_type = f'{self._storage_type}/storage'
        if enc.cek is None:
            raise EncryptionError('CEK is empty while storage initialization')
        try:
            js = {'cek': sirius_sdk.encryption.bytes_to_b58(enc.cek), 'jwe': enc.jwe.as_json()}
            await sirius_sdk.NonSecrets.add_wallet_record(
                type_=storage_type, id_=storage_id, value=json.dumps(js)
            )
        except:
            raise DataVaultStateError('Error while initialize storage encryption')

    async def __create_resource(self, uri: str, is_stream: bool, meta: dict = None, chunk_size: int = None, **attributes):
        self.auth.validate(can_create=True)
        storage = await self._mounted()
        info = await self.__load_resource_info(uri)
        if await storage.exists(uri):
            raise DataVaultCreateResourceError(f'Resource with uri: "{uri}" already exists')
        if info is not None:
            await self.__remove_resource(uri, only_infos=True)
        meta = meta or {}
        if self.META_CREATED_ATTR not in meta:
            meta[self.META_CREATED_ATTR] = datetime.datetime.utcnow().isoformat(sep=' ') + 'Z'
        tags = self.__clean_income_attributes(**attributes)
        tags.update(
            {
                self.ATTR_IS_STREAM: 'yes' if is_stream is True else 'no',
                self.ATTR_CHUNK_SIZE: chunk_size or 'null',
                self.ATTR_CHUNKS_NUM: '0',
                self.ATTR_URN: 'urn:uuid:' + uuid.uuid4().hex
            }
        )
        if is_stream is True:
            tags[self.ATTR_CHUNKS_NUM] = '0'
        await sirius_sdk.NonSecrets.add_wallet_record(
            type_=self._storage_type, id_=uri, value=json.dumps(meta), tags=tags
        )
        try:
            await storage.create(uri)
        except BaseConfidentialStorageError:
            await sirius_sdk.NonSecrets.delete_wallet_record(type_=self._storage_type, id_=uri)
            raise

    async def __remove_resource(self, uri: str, only_infos: bool = True):
        self.auth.validate(can_create=True)
        if only_infos is False:
            storage = await self._mounted()
            await storage.remove(uri)
        info = await self.__load_resource_info(uri)
        if info:
            await sirius_sdk.NonSecrets.delete_wallet_record(type_=self._storage_type, id_=uri)
        self.__clean_caches()

    def __clean_caches(self):
        self.__cached_info = ('', {})
